from dataclasses import dataclass
from typing import List, Protocol
import djelm.codegen.compiler as Compiler
import djelm.codegen.expression as Exp
import djelm.codegen.module_name as Mod
from djelm.codegen.pattern import Pattern, VarPattern


class Writer(Protocol):
    def write(self, indent: int = 0) -> str: ...

    def writeIndented(self, indent: int = 0) -> str:
        if indent == 0:
            return ""
        repeated = ""

        for _ in range(indent):
            repeated += " "
        return repeated


@dataclass(slots=True)
class Spaced(Writer):
    items: List[Writer]

    def write(self, indent: int = 0) -> str:
        return " ".join([w.write(indent) for w in self.items])


@dataclass(slots=True)
class String(Writer):
    val: str

    def write(self, indent: int = 0) -> str:
        return self.val


@dataclass(slots=True)
class Literal(Writer):
    val: str

    def write(self, indent: int = 0) -> str:
        return f'"{ self.val }"'


@dataclass(slots=True)
class Paren(Writer):
    writer: Writer

    def write(self, indent: int = 0) -> str:
        return f"({self.writer.write()})"


@dataclass(slots=True)
class Sep(Writer):
    new_line: bool
    separators: tuple[str, str, str]
    items: List[Writer]

    def write(self, indent: int = 0) -> str:
        pre, sep, post = self.separators

        if post != "":
            post = self.writeIndented(indent) + post

        if self.new_line:
            post = "\n" + post
            separator = "\n" + self.writeIndented(indent) + sep
        else:
            separator = sep

        return f"{pre}{separator.join([w.write() for w in self.items])}{post}"


@dataclass(slots=True)
class SepInline(Writer):
    separators: tuple[str, str, str]
    items: List[Writer]

    def write(self, indent: int = 0) -> str:
        pre, sep, post = self.separators

        return f"{pre}{sep.join([w.write() for w in self.items])}{post}"


@dataclass(slots=True)
class Breaked(Writer):
    writers: List[Writer]

    def write(self, indent: int = 0) -> str:
        return "\n".join([w.write() for w in self.writers])


@dataclass(slots=True)
class Indent(Writer):
    indent: int
    writer: Writer

    def write(self, indent: int = 0) -> str:
        return f"{self.writeIndented(self.indent)}{self.writer.write(self.indent)}"


@dataclass(slots=True)
class Joined(Writer):
    writer: List[Writer]

    def write(self, indent: int = 0) -> str:
        return "".join([w.write() for w in self.writer])


def string(val: str) -> Writer:
    return String(val)


def literal(val: str) -> Writer:
    return Literal(val)


def spaced(items: List[Writer]) -> Writer:
    return Spaced(items)


def paren(writer: Writer) -> Writer:
    return Paren(writer)


def breaked(writers: List[Writer]) -> Writer:
    return Breaked(writers)


def indent(indent: int, writer: Writer) -> Writer:
    return Indent(indent, writer)


def multiFieldRecord(new_line: bool, items: List[Writer]) -> Writer:
    return Sep(new_line, ("{ ", ", ", "}"), items)


def bracesComma(new_line: bool, items: List[Writer]) -> Writer:
    return Sep(new_line, ("{", ", ", "}"), items)


def bracketsComma(new_line: bool, items: List[Writer]) -> Writer:
    return Sep(new_line, ("[", ", ", "]"), items)


def singleRecordField(items: List[Writer]) -> Writer:
    return SepInline(("{ ", ", ", " }"), items)


def sepBy(separators: tuple[str, str, str], newlines: bool, writers: list[Writer]):
    return Sep(newlines, separators, writers)


def join(writers: list[Writer]):
    return Joined(writers)


def epsilon():
    return string("")


def parensIfContainsSpaces(writer: Writer) -> Writer:
    if " " in writer.write():
        return paren(writer)
    else:
        return writer


def writeTypeAnnotation(anno: Compiler.TypeAnnotation) -> Writer:
    match anno:
        case Compiler.Typed(name=name, args=args):
            spaced_writer = Spaced(
                [
                    string(name),
                    *[parensIfContainsSpaces(writeTypeAnnotation(w)) for w in args],
                ]
            )

            return spaced_writer

        case Compiler.Generic(value=value):
            return string(value)

        case Compiler.Record(fields=fields):
            writer_fields = []
            for field in fields:
                writer_fields.append(
                    spaced(
                        [
                            string(field[0]),
                            string(":"),
                            writeTypeAnnotation(field[1].annotation),
                        ]
                    )
                )
            if 1 < len(writer_fields):
                return multiFieldRecord(True, writer_fields)
            else:
                return singleRecordField(writer_fields)

        case _:
            raise Exception("Can't handle that type of annotation")


def writeVariantConstructors(variant: Compiler.Variant) -> Writer:
    writers: list[Writer] = [string(variant.name)]
    writers.extend(
        [
            parensIfContainsSpaces(writeTypeAnnotation(v.annotation))
            for v in variant.annotations
        ]
    )
    return spaced(writers)


def writeModuleName(module: Mod.ModuleName) -> Writer:
    return string(".".join(module.names))


def writePattern(pattern: Pattern) -> Writer:
    match pattern:
        case VarPattern(value=v):
            return string(v)
        case _:
            raise Exception("Cant handle that type of pattern")


def writeExpression(expression: Compiler.Expression) -> Writer:
    match expression:
        case (
            Exp.OperatorApplication(
                symbol=symbol,
                infix_direction=direction,
                left=left,
                right=right,
                range=_,
            ) as op
        ):
            op_range = op.get_range()
            right_range = right.get_range()
            left_range = left.get_range()
            break_right = 0 < right_range.row
            break_left = 0 < left_range.row

            if break_right:
                right_writer = indent(
                    4 + op_range.column,
                    spaced(
                        [
                            string(symbol),
                            writeExpression(right),
                        ]
                    ),
                )
            else:
                right_writer = spaced(
                    [
                        string(symbol),
                        writeExpression(right),
                    ]
                )

            if break_left:
                left_writer = indent(4 + op_range.column, writeExpression(left))
            else:
                left_writer = writeExpression(left)

            match direction:
                case "Left":
                    if break_right or break_left:
                        return breaked(
                            [
                                left_writer,
                                right_writer,
                            ]
                        )
                    else:
                        return spaced(
                            [
                                writeExpression(left),
                                string(symbol),
                                writeExpression(right),
                            ]
                        )

                case "Right":
                    return spaced(
                        [
                            writeExpression(left),
                            writeExpression(right),
                            string(symbol),
                        ]
                    )
                case "Non":
                    return spaced(
                        [
                            writeExpression(left),
                            string(symbol),
                            writeExpression(right),
                        ]
                    )
        case Exp.FunctionOrValue(moduleName=module, name=name):
            match module.names:
                case []:
                    return string(name)
                case _:
                    return join([writeModuleName(module), string("."), string(name)])
        case Exp.Int(value=v, range=name):
            return string(str(v))
        case Exp.Literal(value=v, range=name):
            return literal(v)
        case Exp.Application(expressions=exp):
            match exp:
                case []:
                    return epsilon()
                case [x]:
                    return writeExpression(x)
                case [x, *x_rest]:
                    return spaced(
                        [
                            writeExpression(x),
                            sepBy(
                                ("", " ", ""),
                                False,
                                [writeExpression(ex) for ex in x_rest],
                            ),
                        ]
                    )
                case _:
                    raise Exception("Can't handle that type of espression")

        case Exp.Parenthesized(expression=e):
            return paren(writeExpression(e))

        case Exp.List(members=members, range=_):
            return bracketsComma(False, [writeExpression(m) for m in members])
        case Exp.Lambda(args=args, expression=ex):
            return spaced(
                [
                    join([string("\\"), spaced([writePattern(pat) for pat in args])]),
                    string("->"),
                    writeExpression(ex),
                ]
            )
        case Exp.IfBlock(condition=cond, then_case=then, else_case=els):
            return spaced(
                [
                    string("if"),
                    writeExpression(cond),
                    string("then"),
                    writeExpression(then),
                    string("else"),
                    writeExpression(els),
                ]
            )

        case _:
            raise Exception("Can't handle that expression type")


def writeDeclartion(declaration: Compiler.Declaration) -> Writer:
    """Writer for top level declarations"""
    match declaration.kind:
        case Compiler.AliasDeclaration(name=name, anno=anno):
            return breaked(
                [
                    spaced(
                        [
                            string("type"),
                            string("alias"),
                            string(name),
                            string("="),
                        ]
                    ),
                    indent(4, writeTypeAnnotation(anno.annotation)),
                ]
            )
        case Compiler.CustomTypeDeclaration(name=name, variants=variants):
            return breaked(
                [
                    spaced([string("type"), string(name)]),
                    indent(
                        4,
                        # TODO rstrip string to remove trailing spaces
                        sepBy(
                            ("= ", "| ", ""),
                            True,
                            [writeVariantConstructors(v) for v in variants],
                        ),
                    ),
                ]
            )
        case Compiler.FunctionDeclaration(name=name, signature=sig, expression=exp):
            return breaked(
                [
                    writeSignature(sig),
                    spaced(
                        [
                            string(name),
                            string("="),
                        ]
                    ),
                    indent(4, writeExpression(exp)),
                ]
            )
        case _:
            raise Exception("Cant handle that type of annotation")


def writeSignature(sig: Compiler.Signature) -> Writer:
    return spaced(
        [string(sig.name), string(":"), writeTypeAnnotation(sig.type_annotation)]
    )



from dataclasses import dataclass
from django.conf import settings
import sys
from typing import Generic, TypeVar

from djelm.effect import ExitFailure, ExitSuccess
from djelm.subprocess import SubProcess


T = TypeVar("T")


@dataclass(slots=True)
class CookieCutter(Generic[T]):
    file_dir: str
    output_dir: str
    cookie_template_name: str
    extra: T
    overwrite: bool = False

    def cut(self, logger) -> ExitSuccess[str] | ExitFailure[None, Exception]:
        try:
            from cookiecutter.main import cookiecutter
        except (ImportError, ModuleNotFoundError):
            logger.write("Couldn't find cookie cutter, installing...")
            SubProcess(
                [sys.executable, "-m", "pip", "install", "cookiecutter"],
                settings.BASE_DIR,
            ).open()
            from cookiecutter.main import cookiecutter
        try:
            app_path = cookiecutter(
                self.file_dir,
                output_dir=self.output_dir,
                directory=self.cookie_template_name,
                no_input=True,
                overwrite_if_exists=self.overwrite,
                extra_context=self.extra,
            )
            return ExitSuccess(app_path)
        except Exception as err:
            return ExitFailure(None, err)



from django import template
from ..flags.{{cookiecutter.tag_file}} import key, {{cookiecutter.program_name}}Flags

register = template.Library()


@register.inclusion_tag("djelm/program.html", takes_context=True)
def render_{{ cookiecutter.tag_file }}(context):
    return {"key": key, "flags": {{cookiecutter.program_name}}Flags.parse(0)}


@register.inclusion_tag("djelm/include.html")
def include_{{ cookiecutter.tag_file }}():
    # Generates the script tag for the {{cookiecutter.program_name}}.elm program
    return {"djelm_program": "dist/{{ cookiecutter.program_name }}.js"}



import typing

from djelm.flags.form.primitives import ModelChoiceFieldFlag
from djelm.flags.primitives import Flag

WIDGET_NAMES_T = typing.Union[
    typing.Literal["ModelChoiceField"], typing.Literal["ModelMultipleChoiceField"]
]


WIDGET_NAMES: list[WIDGET_NAMES_T] = [
    "ModelChoiceField",
    "ModelMultipleChoiceField",
]

WIDGET_NAME_TO_DEFAULT_FLAG: dict[WIDGET_NAMES_T, type[Flag]] = {
    "ModelChoiceField": ModelChoiceFieldFlag,
    "ModelMultipleChoiceField": ModelChoiceFieldFlag,
}



from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

T = TypeVar("T")
E = TypeVar("E")


@dataclass(slots=True)
class ExitSuccess(Generic[T]):
    value: T
    tag: Literal["Success"] = "Success"


@dataclass(slots=True)
class ExitFailure(Generic[T, E]):
    meta: T
    err: E
    tag: Literal["Failure"] = "Failure"



from dataclasses import dataclass
import typing as Typ
import djelm.codegen.module_name as Mod
from djelm.codegen.pattern import Pattern
import djelm.codegen.range as Range
import djelm.codegen.compiler as Compiler


@dataclass(slots=True)
class OperatorApplication(Compiler.Expression):
    """Operators are constructed bottom up or right to left"""

    symbol: str
    infix_direction: Typ.Literal["Left"] | Typ.Literal["Right"] | Typ.Literal["Non"]
    left: Compiler.Expression
    right: Compiler.Expression
    range: Range.Range | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng
        self.left.set_range(rng)

    def set_range_column(self, start: int) -> None:
        """
        Set the start column for an operator application

        This trickles in to the left hand side of the operator and allows
        for correct indentation when lines are "breaked"

        e.g.
                someFunc =
                    hello
                        |> world
                    ^   ^
                    |   |
                    |   start column + indentation
                    start column

        """
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range
        self.left.set_range_column(start)

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class FunctionOrValue(Compiler.Expression):
    moduleName: Mod.ModuleName
    name: str
    range: Range.Range | None
    annotation: Compiler.Annotation | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        if self.annotation:
            return self.annotation.annotation
        else:
            return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class Int(Compiler.Expression):
    value: int
    range: Range.Range | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return Compiler.Typed("Int", [])

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class List(Compiler.Expression):
    members: list[Compiler.Expression]
    range: Range.Range | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        match self.members:
            case []:
                return Compiler.Typed("List", [Compiler.Generic("a")])
            case [head, _]:
                head_anno: Compiler.TypeAnnotation | None = head.annotation_type()
                if head_anno:
                    return Compiler.Typed("List", [head_anno])
                else:
                    return Compiler.Typed("List", [Compiler.Generic("a")])
            case _:
                return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class Literal(Compiler.Expression):
    value: str
    range: Range.Range | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return Compiler.Typed("String", [])

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class Application(Compiler.Expression):
    expressions: list[Compiler.Expression]
    annotation: Compiler.TypeAnnotation | None
    range: Range.Range | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class Parenthesized(Compiler.Expression):
    expression: Compiler.Expression
    range: Range.Range | None

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class Lambda(Compiler.Expression):
    args: list[Pattern]
    expression: Compiler.Expression

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)


@dataclass(slots=True)
class IfBlock(Compiler.Expression):
    condition: Compiler.Expression
    then_case: Compiler.Expression
    else_case: Compiler.Expression

    def annotation_type(self) -> Compiler.TypeAnnotation | None:
        return None

    def set_range(self, rng: Range.Range) -> None:
        self.range = rng

    def set_range_column(self, start: int) -> None:
        new_range = self.get_range()
        new_range.column = start
        self.range = new_range

    def get_range(self) -> Range.Range:
        return self.range or Range.Range(0, 0)



import subprocess
import sys

from djelm import get_config
from djelm.subprocess import SubProcess

from .effect import ExitFailure, ExitSuccess


class ElmError(Exception):
    pass


class Elm:
    elm_bin_path: str

    def __init__(self):
        binary = self.elm_binary()
        if get_config("ELM_BIN_PATH"):
            self.elm_bin_path = get_config("ELM_BIN_PATH")
        elif binary.tag == "Success":
            self.elm_bin_path = binary.value
        else:
            raise binary.err

    def command(
        self, args: list[str], target_dir: str
    ) -> ExitSuccess[str] | ExitFailure[None, ElmError | SystemExit]:
        try:
            process = SubProcess([self.elm_bin_path, *list(args)], target_dir, True)
            process.open()
            return ExitSuccess(f"cmd: {args}")
        except subprocess.CalledProcessError:
            sys.exit(1)
        except OSError as err:
            return ExitFailure(
                None,
                ElmError(_elm_binary_log(err)),
            )

    @staticmethod
    def elm_binary() -> ExitSuccess[str] | ExitFailure:
        try:
            # TODO: make platform specific 'platform.system() for Windows'
            p = subprocess.check_output(["which", "elm"])
            return ExitSuccess(p.decode("utf-8"))
        except subprocess.CalledProcessError as err:
            return ExitFailure[None, ElmError](None, ElmError(_elm_binary_log(err)))


def _elm_binary_log(err):
    return (
        f"I can't find an 'elm' binary in your PATH.\n Go to https://guide.elm-lang.org/install/elm.html for "
        f"instructions"
        f"on how to install elm.\n {err}"
    )



from typing import Iterator
import djelm.codegen.expression as Exp
import djelm.codegen.compiler as Compiler


def equals(
    left: Compiler.Expression, right: Compiler.Expression
) -> Compiler.Expression:
    return Exp.OperatorApplication("==", "Left", left, right, None)


def plus(left: Compiler.Expression, right: Compiler.Expression) -> Compiler.Expression:
    return Exp.OperatorApplication("+", "Left", left, right, None)


def pipe(right: Compiler.Expression, left: Compiler.Expression) -> Compiler.Expression:
    """Construct a pipe operator"""
    return Exp.OperatorApplication("|>", "Left", left, right, None)


def pipes(
    top_pipe: Compiler.Expression, other_pipes: Iterator[Compiler.Expression]
) -> Compiler.Expression:
    """
    Supply an iterator of expressions to build a pipe expression bottom to top

    top_pipe = Should be the top most pipe in the chain.
    e.g.
            Decode.succeed Something
            ^   |> required "something" Decode.string
            |
            top_pipe

    other_pipes = The pipe tree will be constructed bottom to top so the iterator should reflect
                  that order
    e.g.
            pipes(top_pipe, iter[1, 2, 3])

            ==
                top_pipe
                   |> 3
                   |> 2
                   |> 1
    """
    try:
        right_pipe = next(other_pipes)
    except StopIteration:
        return top_pipe

    return pipe(right=right_pipe, left=pipes(top_pipe, other_pipes))



from django import template
from ..flags.widgets.{{cookiecutter.tag_file}} import key, {{cookiecutter.program_name}}Flags

register = template.Library()


@register.inclusion_tag("djelm/program.html", takes_context=True, name="render_{{cookiecutter.program_name}}Widget")
def render_{{ cookiecutter.tag_file }}(context):
    return {"key": key, "flags": {{cookiecutter.program_name}}Flags.parse(context["field"])}


@register.inclusion_tag("djelm/include.html", name="include_{{cookiecutter.program_name}}Widget")
def include_{{ cookiecutter.tag_file }}():
    # Generates the script tag for the Widgets/{{cookiecutter.program_name}}.elm program
    return {"djelm_program": "dist/Widgets.{{ cookiecutter.program_name }}.js"}



from typing import Literal
import os
from django.conf import settings
from typing_extensions import TypedDict
from djelm.forms.widgets.main import WIDGET_NAMES, WIDGET_NAMES_T
from .effect import ExitFailure, ExitSuccess
from .utils import (
    get_app_path,
    is_create,
    is_djelm,
    is_init,
    is_program,
    to_program_namespace,
    walk_level,
)

Create = TypedDict("Create", {"command": Literal["create"], "app_name": str})
List = TypedDict("List", {"command": Literal["list"]})
ListWidgets = TypedDict("ListWidgets", {"command": Literal["listwidgets"]})
FindPrograms = TypedDict(
    "FindPrograms", {"command": Literal["findprograms"], "app_name": str}
)
AddProgram = TypedDict(
    "AddProgram",
    {"command": Literal["addprogram"], "app_name": str, "program_name": str},
)
AddProgramHandlers = TypedDict(
    "AddProgramHandlers",
    {
        "command": Literal["addprogramhandlers"],
        "app_name": str,
        "program_name": str,
    },
)
Npm = TypedDict(
    "Npm",
    {"command": Literal["npm"], "app_name": str, "args": list[str]},
)
Elm = TypedDict(
    "Elm",
    {"command": Literal["elm"], "app_name": str, "args": list[str]},
)
Watch = TypedDict(
    "Watch",
    {"command": Literal["watch"], "app_name": str},
)
GenerateModel = TypedDict(
    "GenerateModel",
    {"command": Literal["generatemodel"], "app_name": str, "program_name": str},
)
GenerateModels = TypedDict(
    "GenerateModels",
    {"command": Literal["generatemodels"], "app_name": str},
)
Compile = TypedDict(
    "Compile",
    {"command": Literal["compile"], "app_name": str, "build": bool},
)
AddWidget = TypedDict(
    "AddWidget",
    {"command": Literal["addwidget"], "app_name": str, "widget": WIDGET_NAMES_T},
)


class ValidationError(Exception):
    pass


class Validations:
    def acceptable_command(
        self, labels: list[str], *_
    ) -> (
        ExitSuccess[
            Create
            | List
            | ListWidgets
            | FindPrograms
            | AddProgram
            | AddProgramHandlers
            | Npm
            | Elm
            | Watch
            | GenerateModel
            | GenerateModels
            | Compile
            | AddWidget
        ]
        | ExitFailure[list[str], ValidationError]
    ):
        try:
            self.__check_command_verb(labels[0])
            self.__check_command_combos(labels)
            self.__check_existing(labels)
            return self.__command_exit(labels)
        except ValidationError as err:
            return ExitFailure(labels, err)

    def __check_existing(self, xs: list[str]) -> None:
        match xs:
            case ["create", app_name]:
                app_path_exit = get_app_path(app_name)

                if app_path_exit.tag == "Success" and is_djelm(
                    next(walk_level(app_path_exit.value))[2]
                ):
                    raise ValidationError(self.__app_exists("create", app_name))

                if app_name in settings.INSTALLED_APPS:
                    raise ValidationError(
                        Validations.__not_a_djelm_app("create", app_name)
                    )
            case ["npm", app_name, *_]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("npm", app_name)
                )

                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        Validations.__not_a_djelm_app("npm", app_name)
                    )
            case ["elm", app_name, *_]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("elm", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        Validations.__not_a_djelm_app("elm", app_name)
                    )
            case ["watch", app_name]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("watch", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(self.__not_a_djelm_app("watch", app_name))
            case ["compile", app_name]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("compile", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(self.__not_a_djelm_app("compile", app_name))
            case ["compilebuild", app_name]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("compilebuild", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        self.__not_a_djelm_app("compilebuild", app_name)
                    )
            case ["addprogram", app_name]:
                raise ValidationError(
                    f"""

{Validations.__missing_program_name("addprogram", app_name)}

\033[4m\033[1mHint\033[0m: It helps if the name you choose describes what the program will do, something like \033[1mAddressPicker\033[0m or \033[1mImageCarousel\033[0m.

            e.g. \033[93mdjelm addprogram {app_name} ImageCarousel\033[0m"""
                )
            case ["addprogramhandlers", app_name]:
                raise ValidationError(
                    Validations.__missing_program_name("addprogramhandlers", app_name)
                )

            case ["generatemodel", app_name]:
                raise ValidationError(
                    Validations.__missing_program_name("generatemodel", app_name)
                )

            case ["addprogram", app_name, _]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("addprogram", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        f'{Validations.__not_a_djelm_app("addprogram", app_name)}\n'
                    )
                if not is_init(app_name) or not is_create(app_name):
                    raise ValidationError(
                        f"""
                        {Validations.__missing_files_directories(
                            "addprogram",
                            app_name,
                            [
                                os.path.join(app_name, "static_src", "elm.json"),
                                os.path.join(app_name, "templatetags"),
                            ],
                        )}
\033[4m\033[1mHint\033[0m: These files are usually automatically generated for you when you run the \033[1mcreate\033[0m commands."""
                    )
            case ["addprogramhandlers" as cmd, app_name, prog_name]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings(cmd, app_name)
                )
                namespace = to_program_namespace(prog_name.split("."))
                namespace_path, prog_name = namespace
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        f"{Validations.__not_a_djelm_app(cmd, app_name)}\n"
                    )
                    # or not is_program(app_name, namespace)
                if (
                    not is_init(app_name)
                    or not is_create(app_name)
                    or not is_program(app_name, namespace)
                ):
                    raise ValidationError(
                        f"""
                        {Validations.__missing_files_directories(
                            cmd,
                            app_name,
                            [
                                os.path.join(app_name, "static_src", "elm.json"),
                                os.path.join(app_name, "templatetags"),
                                os.path.join(app_name, "static_src", "src", *namespace_path, prog_name + ".elm"),
                            ],
                        )}
\033[4m\033[1mHint\033[0m: These files are usually automatically generated for you when you run the \033[1mcreate\033[0m commands."""
                    )
            case ["findprograms", app_name]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("addprogram", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        f'{Validations.__not_a_djelm_app("findprograms", app_name)}\n'
                    )

            case ["generatemodel", app_name, p]:
                namespace = to_program_namespace(p.split("."))
                namespace_path, prog_name = namespace

                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("generatemodel", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        Validations.__not_a_djelm_app("generatemodel", app_name)
                    )
                if (
                    not is_init(app_name)
                    or not is_create(app_name)
                    or not is_program(app_name, namespace)
                ):
                    raise ValidationError(
                        f"""
                        {Validations.__missing_files_directories(
                            "generatemodel",
                            app_name,
                            [
                                os.path.join(app_name, "static_src", "elm.json"),
                                os.path.join(app_name, "templatetags"),
                                os.path.join(app_name, "static_src", "src", *namespace_path, prog_name + ".elm"),
                            ],
                        )}
\033[4m\033[1mHint\033[0m: These files are usually automatically generated for you when you run the \033[1mcreate, addprogram\033[0m and \033[1maddwidget\033[0m commands."""
                    )
            case ["generatemodels", app_name]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("generatemodels", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        Validations.__not_a_djelm_app("generatemodel", app_name)
                    )
            case ["addwidget", app_name, _]:
                validated_app_path = self._validate_app_path(
                    app_name, self.__not_in_settings("addwidget", app_name)
                )
                if not is_djelm(next(walk_level(validated_app_path))[2]):
                    raise ValidationError(
                        Validations.__not_a_djelm_app("addwidget", app_name)
                    )

    def __check_command_combos(self, xs: list[str]) -> None:
        match xs:
            case ["create", _]:
                return
            case ["watch", _]:
                return
            case ["addprogram", _, _]:
                return
            case ["addprogramhandlers", _, _]:
                return
            case ["addwidget", _, _]:
                return
            case ["generatemodel", _, _]:
                return
            case ["generatemodels", _]:
                return
            case ["list"]:
                return
            case ["listwidgets"]:
                return
            case ["findprograms", _]:
                return
            case ["compile", _]:
                return
            case ["compilebuild", _]:
                return
            case ["list", *rest]:
                raise ValidationError(
                    Validations.__too_many_command_args("list", list(rest))
                )
            case ["listwidgets", *rest]:
                raise ValidationError(
                    Validations.__too_many_command_args("listwidgets", list(rest))
                )
            case [
                "npm",
            ]:
                raise ValidationError(Validations.__missing_app_name("npm"))
            case [
                "compile",
            ]:
                raise ValidationError(Validations.__missing_app_name("compile"))
            case [
                "compilebuild",
            ]:
                raise ValidationError(Validations.__missing_app_name("compilebuild"))
            case ["compile", app_name, *rest]:
                raise ValidationError(
                    Validations.__too_many_command_args(
                        "compile " + app_name, list(rest)
                    )
                )
            case ["compilebuild", app_name, *rest]:
                raise ValidationError(
                    Validations.__too_many_command_args(
                        "compilebuild " + app_name, list(rest)
                    )
                )
            case [
                "elm",
            ]:
                raise ValidationError(Validations.__missing_app_name("elm"))
            case ["findprograms"]:
                raise ValidationError(Validations.__missing_app_name("findprograms"))
            case ["npm", _, *_]:
                return
            case ["elm", _, *_]:
                return
            case ["watch"] as command_verb:
                raise ValidationError(Validations.__missing_app_name(command_verb[0]))

            case ["create"] as command_verb:
                raise ValidationError(Validations.__missing_app_name(command_verb[0]))
            case ["addwidget", app_name]:
                raise ValidationError(
                    Validations.__missing_argument(
                        "addwidget",
                        app_name,
                        "You need to include the name of the widget you want me to add.\n\nRun \033[1mpython manage.py djelm listwidgets\033[0m to find out what type of widget programs I can add for you.",
                    )
                )
            case [
                "addwidget",
            ]:
                raise ValidationError(Validations.__missing_app_name("addwidget"))
            case ["addwidget", app_name, widget_name, *rest]:
                raise ValidationError(
                    Validations.__too_many_command_args(
                        f"addwidget {app_name} {widget_name}", list(rest)
                    )
                )

    @staticmethod
    def __app_exists(cmd_verb: str, app_name: str) -> str:
        return f"""
\033[91m-- APP ALREADY EXISTS ------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

The app you are targeting is:

    \033[93m{app_name}\033[0m

But according to your \033[1mINSTALLED_APPS\033[0m variable \033[1m{app_name}\033[0m already exists and I can't \033[1m'{cmd_verb}'\033[0m an app that has already been created."""

    @staticmethod
    def __missing_argument(cmd_verb: str, app_name: str, hint: str) -> str:
        return f"""

\033[91m-- MISSING ARGUMENT --------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

The app you are targeting is:

    \033[93m{app_name}\033[0m

But I was expecting another argument.

\033[4m\033[1mHint\033[0m: {hint}"""

    @staticmethod
    def __unknown_widget(cmd_verb: str, widget_name: str) -> str:
        return f"""

\033[91m-- UNKNOWN WIDGET ----------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to add this widget:

    \033[93m{widget_name}\033[0m

But I don't recognize the \033[1m{widget_name}\033[0m widget name.

\033[4m\033[1mHint\033[0m: Run \033[1mpython manage.py djelm listwidgets\033[0m to find out what type of widget programs I can add for you."""

    @staticmethod
    def __too_many_command_args(cmd_verb: str, args: list[str]) -> str:
        extra_args = ""
        for arg in args:
            extra_args += f"{arg} "
        return f"""

\033[91m-- TOO MANY ARGUMENTS ------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

You also included these arguments:

    \033[93m{extra_args}\033[0m

But the \033[1m{cmd_verb}\033[0m command doesn't take any extra arguments.

\033[4m\033[1mHint\033[0m: Try just running \033[1mpython manage.py djelm {cmd_verb}\033[0m"""

    @staticmethod
    def __missing_files_directories(
        cmd_verb: str, app_name: str, files: list[str]
    ) -> str:
        missing = ""

        for file in files:
            missing += f"{file}\n    "
        return f"""
\033[91m-- MISSING FILES/DIRECTORIES ------------------------------------------ command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

The app you are targeting is:

    \033[93m{app_name}\033[0m

But the {cmd_verb} command needs these files/directories to be present:

    \033[93m{missing}\033[0m"""

    @staticmethod
    def __missing_program_name(cmd_verb: str, app_name: str) -> str:
        return f"""
\033[91m-- MISSING PROGRAM NAME ----------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

The app you are targeting is:

    \033[93m{app_name}\033[0m

But you haven't included the name of the Elm program you want me to run this command against.
Make sure that you include a program name when running the \033[1m{cmd_verb}\033[0m command."""

    @staticmethod
    def __missing_app_name(cmd_verb: str) -> str:
        return f"""

\033[91m-- MISSING APP NAME --------------------------------------------------- command/{cmd_verb}\033[0m
        
It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

But you haven't included the name of the djelm app you want me to run this command against.

\033[4m\033[1mHint\033[0m: You can list your current djelm apps by running:

        \033[93mmanage.py djelm list\033[0m"""

    @staticmethod
    def __invalid_strategy(cmd_verb: str) -> str:
        return f"""

\033[91m-- INVALID STRATEGY --------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

But I don't recognise that command.

Perhaps you meant one of these?:

        \033[93mcreate
        addprogram
        watch
        npm
        elm
        generatemodel
        list\033[0m"""

    @staticmethod
    def __not_in_settings(cmd_verb: str, app_name: str) -> str:
        return f"""

\033[91m-- MISSING APP -------------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

The app you are targeting is:

    \033[93m{app_name}\033[0m

But I cant find the \033[1m{app_name}\033[0m app.

\033[4m\033[1mHint\033[0m: Make sure that \033[1m{app_name}\033[0m exists in your \033[1mINSTALLED_APPS\033[0m variable in \033[1msettings.py\033[0m"""

    @staticmethod
    def __not_a_djelm_app(cmd_verb: str, app_name: str) -> str:
        return f"""

\033[91m-- NOT A DJELM APP ---------------------------------------------------- command/{cmd_verb}\033[0m

It looks like you are trying to run the command:

    \033[93m{cmd_verb}\033[0m

The app you are targeting is:

    \033[93m{app_name}\033[0m

But \033[1m{app_name}\033[0m doesn't look like a djelm app and I can't run commands on apps I didn't create."""

    @staticmethod
    def __check_command_verb(s: str) -> None:
        if s not in [
            "create",
            "npm",
            "elm",
            "list",
            "listwidgets",
            "findprograms",
            "addprogram",
            "addprogramhandlers",
            "addwidget",
            "watch",
            "generatemodel",
            "generatemodels",
            "compile",
            "compilebuild",
        ]:
            raise ValidationError(Validations.__invalid_strategy(s))

    @staticmethod
    def __command_exit(
        xs: list[str],
    ) -> (
        ExitSuccess[
            Create
            | List
            | ListWidgets
            | FindPrograms
            | AddProgram
            | AddProgramHandlers
            | Npm
            | Elm
            | Watch
            | GenerateModel
            | GenerateModels
            | Compile
            | AddWidget
        ]
        | ExitFailure[list[str], ValidationError]
    ):
        match xs:
            case ["create", v]:
                return ExitSuccess(Create({"command": "create", "app_name": v}))
            case ["watch", v]:
                return ExitSuccess(Watch({"command": "watch", "app_name": v}))
            case ["compile", v]:
                return ExitSuccess(
                    Compile({"command": "compile", "app_name": v, "build": False})
                )
            case ["compilebuild", v]:
                return ExitSuccess(
                    Compile({"command": "compile", "app_name": v, "build": True})
                )
            case ["addprogram", v, p]:
                return ExitSuccess(
                    AddProgram(
                        {"command": "addprogram", "app_name": v, "program_name": p}
                    )
                )
            case ["addprogramhandlers", v, p]:
                return ExitSuccess(
                    AddProgramHandlers(
                        {
                            "command": "addprogramhandlers",
                            "app_name": v,
                            "program_name": p,
                        }
                    )
                )
            case ["generatemodel", v, p]:
                return ExitSuccess(
                    GenerateModel(
                        {"command": "generatemodel", "app_name": v, "program_name": p}
                    )
                )
            case ["generatemodels", v]:
                return ExitSuccess(
                    GenerateModels({"command": "generatemodels", "app_name": v})
                )
            case ["npm", v, *rest]:
                return ExitSuccess(Npm({"command": "npm", "app_name": v, "args": rest}))
            case ["elm", v, *rest]:
                return ExitSuccess(Elm({"command": "elm", "app_name": v, "args": rest}))
            case ["list"]:
                return ExitSuccess(List({"command": "list"}))
            case ["listwidgets"]:
                return ExitSuccess(ListWidgets({"command": "listwidgets"}))
            case ["addwidget", v, widget_name]:
                if widget_name not in WIDGET_NAMES:
                    raise ValidationError(
                        Validations.__unknown_widget("addwidget", widget_name)
                    )
                return ExitSuccess(
                    AddWidget(
                        {"command": "addwidget", "app_name": v, "widget": widget_name}
                    )
                )
            case ["findprograms", v]:
                return ExitSuccess(
                    FindPrograms({"command": "findprograms", "app_name": v})
                )
            case _ as cmds:
                return ExitFailure(
                    cmds,
                    ValidationError(f"\nI can't handle the command arguments {cmds!s}"),
                )

    @staticmethod
    def _validate_app_path(app_name: str, error_string: str) -> str:
        app_path_exit = get_app_path(app_name)

        if not app_path_exit.tag == "Success":
            raise ValidationError(error_string)
        return app_path_exit.value


# class bcolors:
#     HEADER = '\033[95m'
#     OKBLUE = '\033[94m'
#     OKCYAN = '\033[96m'
#     OKGREEN = '\033[92m'
#     WARNING = '\033[93m'
#     FAIL = '\033[91m'
#     ENDC = '\033[0m'
#     BOLD = '\033[1m'
#     UNDERLINE = '\033[4m'
#



from typing_extensions import Annotated
import typing

from pydantic import BeforeValidator, Field, Strict, TypeAdapter


def match_literal(v: str) -> typing.Callable[[str], str]:
    def do_match(v1):
        if v != v1:
            raise Exception(f"#{v} does not match #{v1} for string literal flag")
        else:
            return v1

    return do_match


def string_literal_adapter(v: str):
    return TypeAdapter(Annotated[str, Strict(), BeforeValidator(match_literal(v))])  # type:ignore


def annotated_string_literal(v: str):
    return Annotated[str, Strict(), BeforeValidator(match_literal(v))]


annotated_string = Annotated[str, Strict()]
annotated_int = Annotated[int, Strict()]
annotated_float = Annotated[float, Strict()]
annotated_bool = Annotated[bool, Strict()]
annotated_alias_key = Annotated[str, Field(pattern=r"^[a-z][A-Za-z0-9_]*$")]

StringAdapter = TypeAdapter(annotated_string)
IntAdapter = TypeAdapter(annotated_int)
FloatAdapter = TypeAdapter(annotated_float)
BoolAdapter = TypeAdapter(annotated_bool)



import subprocess
import sys
import threading
from dataclasses import dataclass
from typing import IO


@dataclass(slots=True)
class SubProcess:
    command: list[str]
    cwd: str
    raise_error: bool = False

    def open(self):
        process = subprocess.Popen(
            self.command,
            cwd=self.cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        stderr_data = []

        def read_stream(stream: IO[str], buffer: list[str]):
            while True:
                char = stream.read(1)
                if not char:
                    break
                sys.stdout.write(char)
                sys.stdout.flush()
                if stream == process.stderr:
                    buffer.append(char)

        stdout_thread = threading.Thread(target=read_stream, args=(process.stdout, []))
        stderr_thread = threading.Thread(
            target=read_stream, args=(process.stderr, stderr_data)
        )

        stdout_thread.start()
        stderr_thread.start()

        process.wait()
        stdout_thread.join()
        stderr_thread.join()

        if process.returncode != 0 or (self.raise_error and stderr_data):
            error_message = "".join(stderr_data).strip()
            raise subprocess.CalledProcessError(
                process.returncode, self.command, output=None, stderr=error_message
            )



import asyncio
from functools import lru_cache
import re
import threading
import aiofiles
import os
import shutil
import subprocess
import sys
from dataclasses import dataclass
from itertools import filterfalse
from typing import Coroutine, Iterable, cast
from django.conf import settings
from typing_extensions import TypedDict
from djelm.forms.widgets.main import WIDGET_NAMES, WIDGET_NAMES_T
from djelm.generators import (
    ModelGenerator,
    ModelBuilder,
    ModelChoiceFieldWidgetGenerator,
    ProgramBuilder,
    ProgramGenerator,
    ProgramHandlersBuilder,
    ProgramHandlersGenerator,
    WidgetModelGenerator,
    entrypoint_cookie_cutter,
)
from watchfiles import awatch
from djelm.cookiecutter import CookieCutter
from djelm.subprocess import SubProcess
from .effect import ExitFailure, ExitSuccess
from .elm import Elm
from .npm import NPM
from .utils import (
    DJELM_VERSION,
    STUFF_ENTRYPOINTS,
    STUFF_NAMESPACE,
    get_app_path,
    get_app_src_path,
    is_djelm,
    is_elm_file_string,
    is_ts_file_string,
    module_name,
    program_file,
    scope_name,
    supporting_ts_files,
    to_program_namespace,
    walk_level,
    widget_scope_name,
)
from .validate import Validations

CreateCookieExtra = TypedDict("CreateCookieExtra", {"app_name": str})
ParsedFile = TypedDict(
    "ParsedFile", {"base": str, "file": str, "supporting_ts_files": set[str]}
)


class StrategyError(Exception):
    pass


@lru_cache()
def unsafe_find_programs(
    src_path: str,
) -> list[ParsedFile]:
    """
    Cached result of the FindPrograms._run_helper method.

    WARNING: The list of returned programs may not be representative of the actual programs in the given src_path.
    """
    result = []
    exception = None
    done = threading.Event()

    def wrapper():
        nonlocal result, exception
        try:
            result = asyncio.run(FindProgramsStrategy(src_path)._run_helper(src_path))
        except Exception as e:
            exception = e
        finally:
            done.set()

    thread = threading.Thread(target=wrapper)
    thread.start()
    done.wait()  # Wait for the thread to finish

    if exception:
        raise exception  # Re-raise any exception that occurred in the thread

    parsed_file_results: list[ParsedFile] = []

    if result:
        for f in result:
            if f.tag == "Success":
                parsed_file_results.append(f.value)

    return parsed_file_results


@dataclass(slots=True)
class FindProgramsStrategy:
    """
    Find elm programs inside the 'src' directory.
    """

    app_name: str

    def run(
        self, logger
    ) -> ExitSuccess[list[ParsedFile]] | ExitFailure[None, StrategyError]:
        src_path = get_app_src_path(self.app_name)

        if src_path.tag != "Success":
            raise src_path.err
        if not os.path.isdir(os.path.join(src_path.value, "src")):
            logger.write("No programs found.")
            return ExitSuccess([])
        parsed_file_results = []

        parsed_file_results = asyncio.run(self._run_helper(src_path.value))

        succeeded_files: list[ParsedFile] = []

        if parsed_file_results:
            logger.write("I found the following programs:\n")
            for f in parsed_file_results:
                if f.tag == "Success":
                    succeeded_files.append(f.value)
                    if f.value["file"].endswith(".elm"):
                        logger.write(f"""    \033[1mbase:\033[0m {f.value["base"]}""")
                        logger.write(f"""    \033[1mfile:\033[0m {f.value["file"]}""")
                        logger.write("\n")
                else:
                    logger.write(str(f.err))

        return ExitSuccess(succeeded_files)

    async def _run_helper(self, src_path: str):
        programs_base_dir = os.path.join(src_path, "src")
        ## Organized by directory to files in that directory
        ts_files_lookup: dict[str, list[str]] = {}
        dir_data_src: Iterable[tuple[str, list[str], list[str]]] = walk_level(
            programs_base_dir
        )
        base_src, dirs_src, files_src = next(dir_data_src)
        ts_files_lookup[base_src] = []
        elm_files_widgets: list[tuple[str, list[str]]] = []

        if "Widgets" in dirs_src:
            dir_data_widgets: Iterable[tuple[str, list[str], list[str]]] = walk_level(
                os.path.join(base_src, "Widgets")
            )
            base_widget_src, _, widget_file_src = next(dir_data_widgets)
            ts_files_lookup[base_widget_src] = []
            widget_files = []
            for f in widget_file_src:
                if is_elm_file_string(f):
                    widget_files.append(f)
                if is_ts_file_string(f):
                    ts_files_lookup[base_widget_src].append(f)
            elm_files_widgets.append((base_widget_src, widget_files))

        elm_files: list[str] = []
        for f in files_src:
            if is_elm_file_string(f):
                elm_files.append(f)
            if is_ts_file_string(f):
                ts_files_lookup[base_src].append(f)

        return await self.__parsed_files(
            [(base_src, elm_files), *elm_files_widgets], ts_files_lookup
        )

    async def __parsed_files(
        self,
        elm_file_data: list[tuple[str, list[str]]],
        ts_files_lookup: dict[str, list[str]],
    ):
        coroutines: list[
            Coroutine[
                None, None, ExitSuccess[ParsedFile] | ExitFailure[None, Exception]
            ]
        ] = []
        for base_path, files in elm_file_data:
            coroutine = [
                self.__parse_file(base_path, file, ts_files_lookup) for file in files
            ]
            coroutines.extend(coroutine)

        return await asyncio.gather(*coroutines)

    async def __parse_file(
        self, base_dir: str, file: str, ts_files_lookup: dict[str, list[str]]
    ) -> ExitSuccess[ParsedFile] | ExitFailure[None, Exception]:
        try:
            handle = await aiofiles.open(os.path.join(base_dir, file))
            content = await handle.read()
            await handle.close()
            if re.search(
                r"^(main : Program Value Model Msg)", content, flags=re.MULTILINE
            ):
                program_name = os.path.splitext(file)[0]
                supporting_files = supporting_ts_files(program_name)
                ts_files = ts_files_lookup.get(base_dir, [])
                ret: ParsedFile = ParsedFile(
                    {
                        "base": base_dir,
                        "file": file,
                        "supporting_ts_files": (
                            set(supporting_files).intersection(set(ts_files))
                        ),
                    }
                )
                return ExitSuccess(ret)
            else:
                return ExitFailure(
                    None, Exception("'main : Program Value Model Msg' not found")
                )

        except Exception as err:
            return ExitFailure(None, err)


@dataclass(slots=True)
class AddProgramStrategy:
    """Create a default elm program."""

    app_name: str
    prog_name: str
    handler: ProgramBuilder

    def run(
        self, logger
    ) -> (
        ExitSuccess[None]
        | ExitFailure[None | str, StrategyError | FileNotFoundError | Exception]
    ):
        src_path = get_app_src_path(self.app_name)
        app_path = get_app_path(self.app_name)

        if src_path.tag != "Success":
            raise src_path.err

        if app_path.tag != "Success":
            raise app_path.err

        try:
            os.makedirs(os.path.join(src_path.value, *STUFF_NAMESPACE, "entrypoints"))
        except FileExistsError:
            pass
        except FileNotFoundError as err:
            raise err

        program_cutters = self.handler.cookie_cutters(
            self.app_name, self.prog_name, src_path.value, DJELM_VERSION
        )

        program_ck_effects = [cutter.cut(logger) for cutter in program_cutters]

        for ck_effect in program_ck_effects:
            if ck_effect.tag != "Success":
                raise ck_effect.err

        try:
            for applicator in self.handler.applicators(
                os.path.join(src_path.value, *STUFF_NAMESPACE),
                src_path.value,
                app_path.value,
                self.prog_name,
                self.app_name,
                logger,
            ):
                applicator.apply(logger)
        except OSError as err:
            raise err

        # Generate model
        model_strat = GenerateModelStrategy(
            self.app_name,
            self.prog_name,
            ModelGenerator(),
            from_source=False,
            watch_mode=False,
        )

        model_strat_effect = model_strat.run(logger)

        if model_strat_effect.tag != "Success":
            raise model_strat_effect.err

        logger.write(
            f"""
I created the \033[92m{self.prog_name}.elm\033[0m program for you!

Right now it's just a default little program that you can change to your hearts content.

You can find it here:

\033[92m{os.path.join(src_path.value, "src", program_file(self.prog_name))}\033[0m

Check out <https://github.com/Confidenceman02/django-elm/blob/main/README.md> to find out how to compile it and see it in the browser.

"""
        )
        return ExitSuccess(None)


@dataclass(slots=True)
class AddProgramHandlersStrategy:
    """
    Add JS handlers for a program.

    Adds a <program_name>.handlers.ts module in the base directory of the given program.
    This module is automatically detected and bundled in with the compiled program.
    Handle JS interop via ports or include JS that is relevant to the program.

    Djelm looks for the following named export callbacks:

        'handlePorts' :: ports -> void
         example:        export function handlePorts(ports): void {
                            // handle port logic
                         }
    """

    app_name: str
    prog_name: str
    src_path: ExitSuccess[str] | ExitFailure[None, Exception]
    app_path: ExitSuccess[str] | ExitFailure[None, Exception]
    handler: ProgramHandlersBuilder

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        if self.src_path.tag != "Success":
            raise self.src_path.err
        if self.app_path.tag != "Success":
            raise self.app_path.err

        cutters = self.handler.cookie_cutters(
            parent_dir=self.src_path.value,
            program_name=self.prog_name,
        )

        cutter_effects = [cutter.cut(logger) for cutter in cutters]

        for ck_effect in cutter_effects:
            if ck_effect.tag != "Success":
                raise ck_effect.err

        logger.write(f"""
I created the \033[92m{self.prog_name}.handlers.ts\033[0m module for you!

You can handle any ports defined in the \033[92m{self.prog_name}.elm\033[0m program with the \033[92mhandlePorts\033[0m function.

You can also include any other JS code related to \033[92m{self.prog_name}.elm\033[0m.

For more information on ports or handlers check out the following resources:

    djelm-docs: \033[1m<https://github.com/Confidenceman02/django-elm/tree/main?tab=readme-ov-file#js-interop>\033[0m
    elm-docs: \033[1m<https://guide.elm-lang.org/interop/ports>\033[0m
""")

        return ExitSuccess(None)


@dataclass(slots=True)
class AddWidgetStrategy:
    """Add a djelm widget"""

    app_name: str
    widget_name: WIDGET_NAMES_T
    handler: ProgramBuilder
    no_deps: bool

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        src_path = get_app_src_path(self.app_name)
        app_path = get_app_path(self.app_name)

        if src_path.tag != "Success":
            raise src_path.err

        if app_path.tag != "Success":
            raise app_path.err

        if not self.no_deps:
            install_deps_effect = self.handler.install_elm_deps(src_path.value, logger)
            if install_deps_effect.tag != "Success":
                raise install_deps_effect.err

        try:
            os.makedirs(os.path.join(src_path.value, *STUFF_NAMESPACE, "entrypoints"))
        except FileExistsError:
            pass
        except FileNotFoundError as err:
            raise err

        # Make widgets dir in elm-stuff
        try:
            os.makedirs(os.path.join(src_path.value, *STUFF_NAMESPACE, "widgets"))
        except FileExistsError:
            pass

        # Make flags dirs
        try:
            os.makedirs(os.path.join(app_path.value, "flags", "widgets"))
        except FileExistsError:
            pass

        # Make Widgets dirs
        try:
            os.makedirs(os.path.join(src_path.value, "src", "Widgets", "Models"))
        except FileExistsError:
            pass

        cutters = self.handler.cookie_cutters(
            self.app_name, self.widget_name, src_path.value, DJELM_VERSION
        )

        cutters_effects = [cutter.cut(logger) for cutter in cutters]

        for cutter_effect in cutters_effects:
            if cutter_effect.tag != "Success":
                raise cutter_effect.err

        # Apply template applicators
        try:
            for applicator in self.handler.applicators(
                os.path.join(src_path.value, *STUFF_NAMESPACE, "widgets"),
                src_path.value,
                app_path.value,
                self.widget_name,
                self.app_name,
                logger,
            ):
                applicator.apply(logger)

        except OSError as err:
            raise err

        # Generate model
        model_strat = GenerateModelStrategy(
            self.app_name,
            self.widget_name,
            WidgetModelGenerator(),
            from_source=True,
            watch_mode=False,
        )
        model_strat_effect = model_strat.run(logger)

        if model_strat_effect.tag != "Success":
            raise model_strat_effect.err

        return ExitSuccess(None)


@dataclass(slots=True)
class CompileStrategy:
    """
    Compiles all elm programs inside of a djelm app
    """

    app_name: str
    build: bool = False
    raise_error: bool = True
    use_cache: bool = False

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        src_path = get_app_src_path(self.app_name)
        if src_path.tag == "Success":
            try:
                os.makedirs(os.path.join(src_path.value, *STUFF_ENTRYPOINTS))
            except FileExistsError:
                pass
            except FileNotFoundError as err:
                raise err

            elm_files: list[ParsedFile] = []

            if self.use_cache:
                elm_files = unsafe_find_programs(src_path.value)
            else:
                unsafe_find_programs.cache_clear()
                elm_files = unsafe_find_programs(src_path.value)

            cookiecutters: list[CookieCutter] = []

            for elm_file in elm_files:
                basedir = os.path.basename(elm_file["base"])
                program_name = os.path.splitext(elm_file["file"])[0]

                if basedir == "Widgets":
                    cookiecutters.append(
                        entrypoint_cookie_cutter(
                            base_name="Widgets.",
                            base_path=f"Widgets{os.path.sep}",
                            src_path=src_path.value,
                            program_name=module_name(program_name),
                            scope=widget_scope_name(self.app_name, program_name),
                            view_prefix="widget",
                            imports=self.compile_imports(
                                elm_file["supporting_ts_files"], f"Widgets{os.path.sep}"
                            ),
                            extras=self.compile_extras(elm_file["supporting_ts_files"]),
                        )
                    )
                else:
                    cookiecutters.append(
                        entrypoint_cookie_cutter(
                            base_name="",
                            base_path="",
                            src_path=src_path.value,
                            program_name=module_name(program_name),
                            scope=scope_name(self.app_name, program_name),
                            view_prefix="",
                            imports=self.compile_imports(
                                elm_file["supporting_ts_files"], ""
                            ),
                            extras=self.compile_extras(elm_file["supporting_ts_files"]),
                        )
                    )

            for cutter in cookiecutters:
                cut = cutter.cut(logger)

                if cut.tag == "Failure":
                    raise cut.err

            try:
                COMPILE_PROGRAM = f"""
                "use strict";
                const _core = require("@parcel/core");
                let bundler = new _core.Parcel({{
                  entries: "./{os.path.join(*STUFF_ENTRYPOINTS)}/*.ts",
                  defaultConfig: "@parcel/config-default",
                  mode: {"'production'" if self.build else "'development'"},
                  defaultTargetOptions: {{
                    distDir: "../static/dist",
                    outputFormat: "esmodule",
                  }},
                }});
                async function Main() {{
                  try {{
                    let {{ bundleGraph, buildTime }} = await bundler.run();
                    let bundles = bundleGraph.getBundles();
                    console.log(`✨ Built ${{bundles.length}} bundles in ${{buildTime}}ms!\n`);
                  }} catch (err) {{
                    if (Array.isArray(err.diagnostics)) {{
                      err.diagnostics.forEach((d) => console.error(d.message ? d.message : d));
                    }} else {{
                      console.error(err.diagnostics);
                    }}
                    process.exit(1);
                  }}
                }}
                Main().then(() => process.exit()).catch((err) => {{console.error(err); process.exit(1);}});
                """
                process = SubProcess(
                    ["node", "-e", COMPILE_PROGRAM],
                    os.path.join(src_path.value),
                    self.raise_error,
                )
                try:
                    process.open()
                except Exception as err:
                    if self.raise_error:
                        raise err
                    else:
                        return ExitSuccess(None)
                return ExitSuccess(None)
            except subprocess.CalledProcessError:
                sys.exit(1)
        return ExitFailure(None, StrategyError("Error"))

    def compile_imports(self, files: set[str], base_path: str) -> list[str]:
        imports = []
        for file in files:
            if "handlers.ts" in file:
                imports.append(
                    f"import {{ handlePorts }} from '../../../src/{base_path}{file}'"
                )
        return imports

    def compile_extras(self, files: set[str]) -> list[str]:
        extras = []
        for file in files:
            if "handlers.ts" in file:
                extras.append("handlePorts(app.ports);")
        return extras


@dataclass(slots=True)
class WatchStrategy:
    """
    Sets up the file watcher for the given djelm app.
    When changes occur it will re-compile the elm programs.

    Modules in the flags directory are also monitored and models generated.
    """

    app_name: str

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        compile = CompileStrategy(self.app_name, raise_error=False, use_cache=True)
        src_path = get_app_src_path(self.app_name)
        app_path = get_app_path(self.app_name)

        if src_path.tag != "Success":
            raise src_path.err

        if app_path.tag != "Success":
            raise app_path.err

        shutil.rmtree(os.path.join(src_path.value, ".parcel-cache"), ignore_errors=True)
        try:
            # first pass compile on start of watch
            compile.run(logger)
            return asyncio.run(
                self.watch(
                    app_path.value,
                    src_path.value,
                    [
                        os.path.join(src_path.value, "src"),
                        os.path.join(app_path.value, "flags"),
                    ],
                    logger,
                )
            )
        except KeyboardInterrupt:
            return ExitSuccess(None)

    async def watch(
        self,
        app_path: str,
        src_path: str,
        dir: list[str],
        logger,
    ):
        compile = CompileStrategy(self.app_name, raise_error=False, use_cache=True)
        async for changes in awatch(*dir):
            for change, f in changes:
                # VIM creates a file to check it can create a file, we want to ignore it
                if f.endswith("4913"):
                    continue
                # If flags change generate models
                # TODO Don't generate a model when a flags module is deleted
                # TODO Only generate a model when the flags output is different to what has already been generated.
                # TODO Validate if it's an actual flag file and not some other python file
                if os.path.join(app_path, "flags") in f and f.endswith(".py"):
                    dirname = os.path.basename(os.path.dirname(f))
                    target_dirs = ["widgets", "flags"]
                    # bail early
                    if dirname not in target_dirs:
                        continue
                    filename = os.path.splitext(os.path.basename(f))[0]
                    namespace = ["src"]
                    generator: ModelBuilder = ModelGenerator()
                    if "widgets" == dirname:
                        namespace.append("Widgets")
                        generator = WidgetModelGenerator()
                    program_name = program_file(filename)
                    # TODO validate if program is an actual Elm program and not some other Elm file
                    is_program = os.path.isfile(
                        os.path.join(src_path, *namespace, program_name)
                    )
                    if is_program:
                        logger.write(f"FLAGS CHANGED: {f}")
                        GenerateModelStrategy(
                            self.app_name,
                            module_name(filename),
                            generator,
                            from_source=True,
                            watch_mode=True,
                        ).run(logger)
                    continue
                # VIM for some reason triggers an ADDED(2) event when saving a buffer
                if change == 1:
                    logger.write(f"FILE ADDED: {f}")
                    # recompile
                    comp = asyncio.to_thread(compile.run, logger)
                    await comp
                if change == 2:
                    logger.write(f"FILE MODIFIED: {f}")
                    # recompile
                    comp = asyncio.to_thread(compile.run, logger)
                    await comp

        return ExitSuccess(None)


@dataclass(slots=True)
class NpmStrategy:
    """
    A helper class to call the node package manager binary on the given djelm app.

    The commands passed will execute in the directory of the djelm app that contains the package.json file.
    """

    app_name: str
    args: list[str]

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        npm = NPM()
        src_path = get_app_src_path(self.app_name)

        if src_path.tag != "Success":
            # TODO Better error
            raise src_path.err

        npm_exit = npm.command(src_path.value, self.args)

        if npm_exit.tag != "Success":
            # TODO Better error
            raise npm_exit.err

        logger.write("\033[92mCompleted successfully.\033[0m")
        return ExitSuccess(None)


@dataclass(slots=True)
class ElmStrategy:
    """
    A Helper class to call an elm binary on the given djelm app.

    The commands passed will execute in the directory of the djelm app that contains the elm.json file.
    """

    app_name: str
    args: list[str]

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        elm: Elm = Elm()
        src_path = get_app_src_path(self.app_name)
        if src_path.tag == "Success":
            elm_exit = elm.command(self.args, target_dir=src_path.value)

            if elm_exit.tag == "Success":
                return ExitSuccess(None)
            else:
                return ExitFailure(None, err=StrategyError(elm_exit.err))
        return ExitFailure(None, err=StrategyError())


class GenerateModelStrategy:
    """Generate a model and decoders for an Elm program"""

    def __init__(
        self,
        app_name,
        prog_name,
        handler: ModelBuilder,
        from_source: bool,
        watch_mode: bool,
    ) -> None:
        self.app_name = app_name
        self.prog_name = prog_name
        self.handler = handler
        self.from_source = from_source
        self.watch_mode = watch_mode

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        app_path = get_app_path(self.app_name)
        src_path = get_app_src_path(self.app_name)
        handler = self.handler

        if app_path.tag != "Success":
            return ExitFailure(meta=None, err=StrategyError(app_path.err))
        if src_path.tag != "Success":
            return ExitFailure(meta=None, err=StrategyError(src_path.err))

        flags_effect = handler.load_flags(
            app_path.value, self.prog_name, self.from_source, self.watch_mode, logger
        )

        if flags_effect.tag != "Success":
            return ExitFailure(meta=None, err=StrategyError(flags_effect.err))

        try:
            os.makedirs(os.path.join(src_path.value, *STUFF_NAMESPACE))
        except FileExistsError:
            pass
        except FileNotFoundError as err:
            return ExitFailure(meta=None, err=StrategyError(err))
        ck = handler.cookie_cutter(
            flags_effect.value, self.app_name, self.prog_name, src_path.value
        )
        try:
            cookie_effect = ck.cut(logger)

            if cookie_effect.tag != "Success":
                return ExitFailure(meta=None, err=StrategyError(cookie_effect.err))

            for applicator in handler.applicators(
                cookie_effect.value,
                src_path.value,
                app_path.value,
                self.prog_name,
                self.app_name,
                logger,
            ):
                applicator.apply(logger)

            return ExitSuccess(None)
        except OSError as err:
            return ExitFailure(meta=None, err=StrategyError(err))


@dataclass(slots=True)
class GenerateModelsStrategy:
    """Generate model and decoder for all djelm elm programs"""

    app_name: str

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        app_path = get_app_path(self.app_name)
        src_path = get_app_src_path(self.app_name)

        if app_path.tag != "Success":
            raise app_path.err
        if src_path.tag != "Success":
            raise src_path.err

        unsafe_find_programs.cache_clear()
        all_programs = unsafe_find_programs(src_path.value)

        for program in all_programs:
            baseDir = os.path.basename(program["base"])
            program_name = os.path.splitext(program["file"])[0]
            generator = ModelGenerator()
            if baseDir == "Widgets":
                generator = program_namespace_to_model_builder([baseDir, program_name])

            GenerateModelStrategy(
                self.app_name,
                program_name,
                generator,
                from_source=True,
                watch_mode=False,
            ).run(logger)

        return ExitSuccess(None)


class ListStrategy:
    apps: list[str] = settings.INSTALLED_APPS

    def run(self, logger) -> ExitSuccess[list[str]] | ExitFailure[None, StrategyError]:
        app_path_exits = filterfalse(
            lambda x: x.tag == "Failure", map(get_app_path, self.apps)
        )

        dir_data: Iterable[tuple[str, list[str], list[str]]] = map(
            next,
            map(lambda p: walk_level(p.value), app_path_exits),  # type:ignore
        )

        django_elm_apps = [os.path.basename(r) for r, _, f in dir_data if is_djelm(f)]

        apps = ""

        for app in django_elm_apps:
            apps += f"\033[93m{app}\033[0m\n\t"

        logger.write(
            f"""
Here are all the djelm apps I found:

        {apps}"""
        )
        return ExitSuccess(django_elm_apps)


class ListWidgetsStrategy:
    widgets = WIDGET_NAMES

    def run(
        self, logger
    ) -> ExitSuccess[list[WIDGET_NAMES_T]] | ExitFailure[None, StrategyError]:
        widgets = ""
        for w in self.widgets:
            widgets += f"\033[93m{w}\033[0m\n\t"

        logger.write(
            f"""
Here are all the widgets I support:

        {widgets}"""
        )
        return ExitSuccess(self.widgets)


@dataclass(slots=True)
class CreateStrategy:
    app_name: str

    def run(self, logger) -> ExitSuccess[None] | ExitFailure[None, StrategyError]:
        ck = CookieCutter[CreateCookieExtra](
            file_dir=os.path.dirname(__file__),
            output_dir=os.getcwd(),
            cookie_template_name="project_template",
            extra={
                "app_name": self.app_name.strip(),
            },
        )

        cut_cookie = ck.cut(logger)

        if cut_cookie.tag == "Success":
            app_name = os.path.basename(cut_cookie.value)
            logger.write(
                f"""

Hi there! I have created the \033[92m{f"{app_name}"}\033[0m djelm app for you. This is where all your Elm programs will live.

Now you may be wondering, what will be in this app? Where do I add Elm files?
How does it work with django so I can see it in the browser? How will my code grow? Do I need
more directories?

Check out <https://github.com/Confidenceman02/django-elm/blob/main/README.md> for all the answers!

For now, make sure you add \033[92m"{f"{app_name}"}"\033[0m to the \033[1mINSTALLED_APPS\033[0m variable in \033[1msettings.py\033[0m.
"""
            )
            return ExitSuccess(None)
        return ExitFailure(None, StrategyError(cut_cookie.err))


@dataclass(slots=True)
class Strategy:
    def create(
        self, labels, options: dict[str, bool]
    ) -> (
        ElmStrategy
        | CreateStrategy
        | ListStrategy
        | ListWidgetsStrategy
        | AddProgramStrategy
        | AddProgramHandlersStrategy
        | NpmStrategy
        | WatchStrategy
        | GenerateModelStrategy
        | GenerateModelsStrategy
        | CompileStrategy
        | AddWidgetStrategy
        | FindProgramsStrategy
    ):
        e = Validations().acceptable_command(labels)
        match e:
            case ExitFailure(err=err):
                raise err
            case ExitSuccess(value={"command": "create", "app_name": app_name}):
                return CreateStrategy(cast(str, app_name))
            case ExitSuccess(value={"command": "watch", "app_name": app_name}):
                return WatchStrategy(cast(str, app_name))
            case ExitSuccess(
                value={
                    "command": "addprogram",
                    "app_name": app_name,
                    "program_name": pn,
                }
            ):
                return AddProgramStrategy(
                    cast(str, app_name), cast(str, pn), handler=ProgramGenerator()
                )
            case ExitSuccess(
                value={
                    "command": "addprogramhandlers",
                    "app_name": app_name,
                    "program_name": pn,
                }
            ):
                program_path, program_name = to_program_namespace(pn.split("."))
                return AddProgramHandlersStrategy(
                    cast(str, app_name),
                    program_name,
                    get_app_src_path(app_name),
                    get_app_path(app_name),
                    handler=program_namespace_to_handlers_builder(
                        [*program_path, program_name]
                    ),
                )
            case ExitSuccess(
                value={
                    "command": "generatemodel",
                    "app_name": app_name,
                    "program_name": pn,
                }
            ):
                program_path, program_name = to_program_namespace(pn.split("."))
                return GenerateModelStrategy(
                    cast(str, app_name),
                    program_name,
                    handler=program_namespace_to_model_builder(
                        [*program_path, program_name]
                    ),
                    from_source=True,
                    watch_mode=False,
                )
            case ExitSuccess(value={"command": "generatemodels", "app_name": app_name}):
                return GenerateModelsStrategy(app_name)
            case ExitSuccess(value={"command": "list"}):
                return ListStrategy()
            case ExitSuccess(value={"command": "listwidgets"}):
                return ListWidgetsStrategy()
            case ExitSuccess(value={"command": "findprograms", "app_name": app_name}):
                return FindProgramsStrategy(app_name=app_name)
            case ExitSuccess(
                value={"command": "npm", "app_name": app_name, "args": args}
            ):
                return NpmStrategy(cast(str, app_name), cast(list[str], args))
            case ExitSuccess(
                value={"command": "elm", "app_name": app_name, "args": args}
            ):
                return ElmStrategy(cast(str, app_name), cast(list[str], args))
            case ExitSuccess(
                value={"command": "compile", "app_name": app_name, "build": build}
            ):
                return CompileStrategy(cast(str, app_name), cast(bool, build))
            case ExitSuccess(
                value={"command": "addwidget", "app_name": app_name, "widget": widget}
            ):
                return AddWidgetStrategy(
                    cast(str, app_name),
                    cast(WIDGET_NAMES_T, widget),
                    handler=widget_name_to_program_builder(widget),
                    no_deps=options.get("no_deps", False),
                )
            case _ as x:
                raise StrategyError(f"Unable to handle {x}")


def widget_name_to_program_builder(widget_name: WIDGET_NAMES_T) -> ProgramBuilder:
    match widget_name:
        case "ModelChoiceField":
            return ModelChoiceFieldWidgetGenerator()
        case "ModelMultipleChoiceField":
            return ModelChoiceFieldWidgetGenerator()


def program_namespace_to_handlers_builder(
    namespace: list[str],
) -> ProgramHandlersBuilder:
    match namespace:
        case [_]:
            return ProgramHandlersGenerator(base_path=[], target_dir="src")
        case ["Widgets", widget_name]:
            if widget_name in WIDGET_NAMES:
                return ProgramHandlersGenerator(["src"], "Widgets")
            else:
                raise StrategyError(__unknown_widget("addprogramhandlers", widget_name))
        case [_, *_]:
            raise StrategyError(
                f"""\033[91m-- INVALID PROGRAM ----------------------------- command/addprogramhandlers\033[0m

It looks like you are trying to run the command addprogramhandlers for this program:

    \033[93m{".".join(namespace)}\033[0m

\033[4m\033[1mHint\033[0m: Make sure the program you are targeting exists.
"""
            )
        case _:
            raise StrategyError(
                f"I can't resolve a {ProgramHandlersBuilder.__name__} for {'.'.join(namespace)}"
            )


def program_namespace_to_model_builder(namespace: list[str]) -> ModelBuilder:
    match namespace:
        case [_]:
            return ModelGenerator()
        case ["Widgets", widget_name]:
            if widget_name in WIDGET_NAMES:
                return WidgetModelGenerator()
            else:
                raise StrategyError(__unknown_widget("generatemodel", widget_name))

        case [head, *_]:
            raise StrategyError(
                f"""\033[91m-- INVALID PROGRAM NAMESPACE ----------------------------- command/generatemodel\033[0m

I can't generate a model for programs in the {head} namespace.

\033[4m\033[1mHint\033[0m: Make sure the program you are generating a model for exists in one of the following directories.

\033[1msrc\033[0m - python manage.py djelm generatemodel <app> <program_name>
\033[1mWidgets\033[0m  - python manage.py djelm generatemodel <app> Widgets.<widget_program>
"""
            )
        case _:
            raise StrategyError(f"I can't resolve a ModelBuilder for {namespace}")


def __unknown_widget(command: str, widget_name: str) -> str:
    return f"""

\033[91m-- UNKNOWN WIDGET --------------------------------- command/{command}\033[0m

It looks like you are trying to run the command {command} for this widget:

    \033[93m{widget_name}\033[0m

But I don't recognize the \033[1m{widget_name}\033[0m widget name.

\033[4m\033[1mHint\033[0m: Run \033[1mpython manage.py djelm listwidgets\033[0m to list the type of widget programs I can work with."""



from dataclasses import dataclass
from typing import Annotated, Any
from django.db import models
from django.forms.boundfield import BoundField
from pydantic import TypeAdapter, ValidationError
from pydantic.functional_validators import BeforeValidator
from djelm.flags.form.helpers import ModelChoiceFieldVariant, ModelChoiceFlagHelper
from djelm.flags.primitives import Flag
from django.forms.models import ModelChoiceField


@dataclass(slots=True)
class ModelChoiceFieldFlag(Flag):
    variants: list[type[models.Model]] | None = None

    def anno(self) -> Annotated:
        return Annotated[
            ModelChoiceFlagHelper.object_annotation(self._get_variants()),  # type:ignore
            BeforeValidator(self.serializer),
        ]

    def adapter(self):
        return TypeAdapter(self.anno())

    def obj(self) -> Flag:
        return ModelChoiceFlagHelper.object_flag(self._get_variants())

    def _get_variants(self) -> list[ModelChoiceFieldVariant]:
        if self.variants:
            return [ModelChoiceFieldVariant(mod) for mod in self.variants]
        else:
            return []

    def serializer(self, field: Any):
        """
        Serialize a BoundField to a structure pydantic can validate.
        """
        try:
            assert isinstance(field, BoundField) and isinstance(
                field.field, ModelChoiceField
            )
        except Exception as err:
            raise ValidationError(err)

        return ModelChoiceFlagHelper.field_to_dict(field, self._get_variants())


class ModelMultipleChoiceFieldFlag(ModelChoiceFieldFlag):
    def __init__(self, variants: list[type[models.Model]] | None = None):
        super().__init__(variants)



from typing import List
import djelm.codegen.compiler as Compiler
import djelm.codegen.writer as Writer
import djelm.codegen.format as Format
from djelm.codegen.utils import foldl


def getAliases(anno: Compiler.Annotation) -> dict[str, Compiler.TypeAnnotation]:
    return anno.aliases


def mergeAliases(
    x: dict[str, Compiler.TypeAnnotation], anno: Compiler.Annotation
) -> dict[str, Compiler.TypeAnnotation]:
    z = x.copy()
    z.update(getAliases(anno))
    return z


def toAliasKey(k: str) -> str:
    return f"{k[0].upper()}{k[1:]}"


def addAlias(
    name: str,
    target: Compiler.Annotation,
    alias_cache: dict[str, Compiler.TypeAnnotation],
) -> dict[str, Compiler.TypeAnnotation]:
    aliasKey: str = toAliasKey(name)
    alias_cache.update({f"{aliasKey}": target.annotation})
    return alias_cache


def typed(name: str, args: List[Compiler.Annotation]) -> Compiler.Annotation:
    args_anno: List[Compiler.TypeAnnotation] = []
    for arg in args:
        args_anno.append(arg.annotation)

    return Compiler.Annotation(
        Compiler.Typed(name, args_anno),
        foldl(mergeAliases, {}, args),
    )


def string():
    """Elm String annotation"""
    return typed("String", [])


def int():
    """Elm Int annotation"""
    return typed("Int", [])


def float():
    """Elm Float annotation"""
    return typed("Float", [])


def bool() -> Compiler.Annotation:
    """Elm Bool annotation"""
    return typed("Bool", [])


def maybe(anno: Compiler.Annotation):
    """Elm Maybe annotation"""
    return typed("Maybe", [anno])


def list(anno: Compiler.Annotation):
    """Elm List annotation"""
    return typed("List", [anno])


def alias(name: str, anno: Compiler.Annotation):
    """Elm alias annotation"""
    return Compiler.Annotation(
        Compiler.Typed(Format.alias_type(name), []), addAlias(name, anno, {})
    )


def record(fields: List[tuple[str, Compiler.Annotation]]) -> Compiler.Annotation:
    """Elm Dict annotation"""
    return Compiler.Annotation(
        Compiler.Record(fields),
        foldl(lambda acc, field: mergeAliases(acc, field[1]), {}, fields),
    )


def toString(anno: Compiler.Annotation) -> str:
    return Writer.writeTypeAnnotation(anno.annotation).write()



import os

from django.apps import apps

from djelm.effect import ExitFailure, ExitSuccess
from importlib.metadata import version

DJELM_VERSION = version("djelm")
STUFF_NAMESPACE = ("elm-stuff", f"djelm_{DJELM_VERSION}")
STUFF_ENTRYPOINTS = (*STUFF_NAMESPACE, "entrypoints")


def get_app_path(app_name) -> ExitSuccess[str] | ExitFailure[None, Exception]:
    app_label = app_name.split(".")[0]
    try:
        path = apps.get_app_config(app_label).path
        return ExitSuccess(path)
    except LookupError as err:
        return ExitFailure(meta=None, err=err)


def get_app_src_path(app_name: str) -> ExitSuccess[str] | ExitFailure[None, Exception]:
    path_exit = get_app_path(app_name)
    if path_exit.tag == "Success":
        return ExitSuccess(os.path.join(path_exit.value, "static_src"))
    return path_exit


def install_pip_package(pkg_name: str):
    from pip._internal import main

    main(["install", pkg_name])


def walk_level(some_dir: str, level=1):
    some_dir = some_dir.rstrip(os.path.sep)
    assert os.path.isdir(some_dir)
    num_sep = some_dir.count(os.path.sep)
    for root, dirs, files in os.walk(some_dir):
        yield root, dirs, files
        num_sep_this = root.count(os.path.sep)
        if num_sep + level <= num_sep_this:
            del dirs[:]


def is_djelm(files: list[str]) -> bool:
    test = ".djelm"
    found = False
    for f in files:
        if test in f:
            found = True
            break
    return found


def is_create(app_name: str) -> bool:
    path_exit = get_app_path(app_name)
    if path_exit.tag == "Success":
        tags = os.path.isdir(
            os.path.join(
                path_exit.value,
                "templatetags",
            )
        )
        if tags:
            return True
    return False


def is_init(app_name: str) -> bool:
    path_exit = get_app_path(app_name)
    if path_exit.tag == "Success":
        f = os.path.isfile(
            os.path.join(
                path_exit.value,
                "static_src",
                "elm.json",
            )
        )
        src = os.path.isdir(
            os.path.join(
                path_exit.value,
                "static_src",
                "src",
            )
        )
        return f and src
    return False


def is_program(app_name: str, namespace: tuple[list[str], str]) -> bool:
    namespace_path, prog_name = namespace
    path_exit = get_app_src_path(app_name)
    if path_exit.tag == "Success":
        f = os.path.isfile(
            os.path.join(
                path_exit.value,
                "src",
                *namespace_path,
                prog_name + ".elm",
            )
        )
        return f
    return False


def to_program_namespace(namespace: list[str]) -> tuple[list[str], str]:
    prog_name = namespace.pop()

    return (namespace, prog_name)


def module_name(app_name: str):
    return app_name[0].upper() + app_name[1:]


def view_name(prog_name: str):
    return prog_name[0].upper() + prog_name[1:].lower()


def tag_file_name(prog_name: str):
    return prog_name[0].lower() + prog_name[1:]


def program_file(app_name: str):
    return module_name(app_name) + ".elm"


def scope_name(app_name: str, prog_name: str) -> str:
    return (app_name.lower() + prog_name.lower().replace("_", "")).replace("_", "")


def widget_scope_name(app_name: str, prog_name: str) -> str:
    return ("widgets" + app_name.lower() + prog_name.lower().replace("_", "")).replace(
        "_", ""
    )


def is_elm_file_string(string_file: str) -> bool:
    return string_file.endswith(".elm")


def is_ts_file_string(string_file: str) -> bool:
    return string_file.endswith(".ts")


def supporting_ts_files(prog_name: str) -> list[str]:
    return [prog_name + ".handlers.ts"]



from dataclasses import dataclass
from types import UnionType
from typing import Any, NotRequired, Optional
import typing
from typing_extensions import TypedDict
from django.db import models
from django.forms.boundfield import BoundField, BoundWidget
from django.db.models.fields import CharField, BooleanField, IntegerField, FloatField
from pydantic import BaseModel
from djelm.flags.primitives import (
    BoolFlag,
    CustomTypeFlag,
    Flag,
    IntFlag,
    ListFlag,
    NullableFlag,
    ObjectFlag,
    StringFlag,
    FloatFlag,
)

SUPPORTED_VARIANT_FIELDS = [CharField, BooleanField, IntegerField, FloatField]

DEFAULT_OPTION_TYPES = [
    ("choice_label", str),
    ("value", str),
    ("selected", bool),
]

DEFAULT_OPTION_ROOT_TYPES = [
    ("help_text", str),
    ("auto_id", str),
    ("id_for_label", str),
    ("label", str),
    ("name", str),
    ("widget_type", str),
]

OPTION_BASE_MODEL = type(
    "OptionBaseModel",
    (BaseModel,),
    {"__annotations__": dict(DEFAULT_OPTION_TYPES)},
)

MODEL_CHOICE_FIELD_BASE_MODEL = type(
    "ModelChoiceFieldBaseModelAnno",
    (BaseModel,),
    {
        "__annotations__": dict(
            [*DEFAULT_OPTION_ROOT_TYPES, ("options", list[OPTION_BASE_MODEL])]
        )
    },
)

ModelChoiceFieldObjects = TypedDict(
    "ModelChoiceFieldObjects",
    {
        "fields": list[str],
        "flag": Flag,
        "annotations": dict[
            str, typing.Union[type[str], type[int], type[float], type[bool]] | UnionType
        ],
    },
)

ModelChoiceFieldOptions = TypedDict(
    "ModelChoiceFieldOptions",
    {
        "choice_label": str,
        "value": str,
        "selected": bool,
        "instance": NotRequired[dict[str, Any]],
    },
)

ModelChoiceFieldType = TypedDict(
    "ModelChoiceFieldType",
    {
        "help_text": str,
        "auto_id": str,
        "id_for_label": str,
        "label": Optional[str],
        "name": str,
        "widget_type": str,
        "options": list[ModelChoiceFieldOptions],
    },
)


@dataclass(slots=True)
class ModelChoiceFieldVariant:
    model: type[models.Model]

    def get_field_objects(self) -> ModelChoiceFieldObjects:
        resolved_field_names: list[str] = []
        resolved_flags: list[Flag] = []
        resolved_annotations: list[
            typing.Union[type[str], type[int], type[float], type[bool] | UnionType]
        ] = []
        fields = self.model._meta.get_fields()

        for field in fields:
            if any([isinstance(field, f) for f in SUPPORTED_VARIANT_FIELDS]):
                resolved_field_names.append(field.name)
                resolved_flags.append(self.field_to_flag(field))
                resolved_annotations.append(self.field_to_annotation(field))

        return {
            "fields": resolved_field_names,
            "flag": ObjectFlag(dict(zip(resolved_field_names, resolved_flags))),
            "annotations": dict(zip(resolved_field_names, resolved_annotations)),
        }

    def get_classname(self) -> str:
        return self.model.__name__

    def get_instance_values(self, model: models.Model) -> dict[str, Any]:
        field_names = self.get_field_objects()["fields"]
        resolved_fields = {}

        for field in field_names:
            resolved_fields[field] = getattr(model, field)

        return resolved_fields

    def field_to_flag(self, field) -> Flag:
        is_null: bool = field.null
        if isinstance(field, CharField):
            if is_null:
                return NullableFlag(StringFlag())
            else:
                return StringFlag()
        if isinstance(field, BooleanField):
            if is_null:
                return NullableFlag(BoolFlag())
            else:
                return BoolFlag()
        if isinstance(field, IntegerField):
            if is_null:
                return NullableFlag(IntFlag())
            else:
                return IntFlag()
        if isinstance(field, FloatField):
            if is_null:
                return NullableFlag(FloatFlag())
            else:
                return FloatFlag()
        raise Exception(
            f"The model field type {field.__class__.__name__} is not supported."
        )

    def field_to_annotation(
        self, field
    ) -> typing.Union[type[str], type[int], type[float], type[bool]] | UnionType:
        is_null: bool = field.null
        if isinstance(field, CharField):
            if is_null:
                return typing.Optional[str]
            else:
                return str
        if isinstance(field, BooleanField):
            if is_null:
                return typing.Optional[bool]
            else:
                return bool
        if isinstance(field, IntegerField):
            # TODO Handle AutoField, BigAutoField and BigIntegerField
            if is_null:
                return typing.Optional[int]
            else:
                return int
        if isinstance(field, FloatField):
            if is_null:
                return typing.Optional[float]
            else:
                return float
        raise Exception(
            f"The model field type {field.__class__.__name__} is not supported."
        )


class ModelChoiceFlagHelper:
    @staticmethod
    def field_to_dict(
        field: BoundField, variants: list[ModelChoiceFieldVariant]
    ) -> ModelChoiceFieldType:
        return {
            "help_text": str(field.help_text),
            "auto_id": field.auto_id,
            "id_for_label": field.id_for_label,
            "label": field.label,  # type:ignore
            "name": field.name,
            "widget_type": field.widget_type,
            "options": [
                ModelChoiceFlagHelper.widget_to_options(opt, variants=variants)
                for opt in field.subwidgets
            ],
        }

    @staticmethod
    def widget_to_options(
        option: BoundWidget, variants: list[ModelChoiceFieldVariant]
    ) -> ModelChoiceFieldOptions:
        base_option: ModelChoiceFieldOptions = {
            "choice_label": option.choice_label,
            "value": str(option.data["value"]),
            "selected": option.data["selected"],
        }

        if variants:
            bound_instance = option.data["value"].instance

            assert isinstance(bound_instance, models.Model)

            variant = None
            for v in variants:
                if isinstance(bound_instance, v.model):
                    variant = v

            if variant:
                base_option["instance"] = variant.get_instance_values(bound_instance)

        return base_option

    @staticmethod
    def object_annotation(variants: list[ModelChoiceFieldVariant]):
        """Annotations for the ModelChoiceField field"""

        variant_annotations = []

        if variants:
            for variant in variants:
                instance_root = [*DEFAULT_OPTION_ROOT_TYPES]
                option_with_instance = [*DEFAULT_OPTION_TYPES]
                option_with_instance.append(
                    (
                        "instance",
                        type(
                            variant.get_classname() + "InstanceBaseModel",
                            (BaseModel,),
                            {
                                "__annotations__": variant.get_field_objects()[
                                    "annotations"
                                ]
                            },
                        ),
                    )
                )
                instance_root.append(
                    (
                        "options",
                        list[
                            type(
                                variant.get_classname() + "BaseModel",
                                (BaseModel,),
                                {"__annotations__": dict(option_with_instance)},
                            )
                        ],
                    )
                )

                variant_annotations.append(
                    type(
                        variant.get_classname() + "VariantBaseModel",
                        (BaseModel,),
                        {"__annotations__": dict(instance_root)},
                    )  # type:ignore
                )

        if variant_annotations:
            return typing.Union[*variant_annotations]  # type:ignore
        else:
            return MODEL_CHOICE_FIELD_BASE_MODEL

    @staticmethod
    def object_flag(variants: list[ModelChoiceFieldVariant] | None = None) -> Flag:
        """A Flag analogue for the ModelChoiceField field"""

        # The default option when no variants exist
        resolved_option: Flag = ObjectFlag(
            {
                "choice_label": StringFlag(),
                "value": StringFlag(),
                "selected": BoolFlag(),
            }
        )

        if variants:
            options: list[tuple[str, Flag]] = [
                (
                    variant.get_classname(),
                    ObjectFlag(
                        {
                            "choice_label": StringFlag(),
                            "value": StringFlag(),
                            "selected": BoolFlag(),
                            "instance": variant.get_field_objects()["flag"],
                        }
                    ),
                )
                for variant in variants
            ]

            resolved_option = CustomTypeFlag(options)

        return ObjectFlag(
            {
                "help_text": StringFlag(),
                "auto_id": StringFlag(),
                "id_for_label": StringFlag(),
                "label": NullableFlag(StringFlag()),
                "name": StringFlag(),
                "widget_type": StringFlag(),
                "options": ListFlag(resolved_option),
            }
        )



from django.conf import settings


def get_config(setting_name) -> str:
    return {
        "NODE_PACKAGE_MANAGER": getattr(settings, "NODE_PACKAGE_MANAGER", "pnpm"),
        "ELM_BIN_PATH": getattr(settings, "ELM_BIN_PATH", "elm"),
    }[setting_name]



import os

app_name = "{{cookiecutter.app_name}}"

REMOVE_PATHS = [
    "static/dist/.gitkeep",
    "templates/.gitkeep",
    "templates/{{cookiecutter.app_name}}/.gitkeep",
    "static_src/src/Models/.gitkeep",
    "flags/.gitkeep",
]

for path in REMOVE_PATHS:
    path = path.strip()
    if path and os.path.exists(path):
        os.unlink(path) if os.path.isfile(path) else os.rmdir(path)



import os
import shutil
from typing import Protocol, Sequence, cast
from typing_extensions import TypedDict

from djelm.cookiecutter import CookieCutter
from djelm.effect import ExitFailure, ExitSuccess
from djelm.flags import Flags
from djelm.flags.primitives import IntFlag
from djelm.forms.widgets.main import WIDGET_NAME_TO_DEFAULT_FLAG, WIDGET_NAMES_T
from djelm.npm import NPM, NPMError
import djelm.flag_loader as FlagLoader
from djelm.utils import (
    STUFF_NAMESPACE,
    module_name,
    scope_name,
    tag_file_name,
    view_name,
    widget_scope_name,
)

#  _____           _                  _
# |  __ \         | |                | |
# | |__) | __ ___ | |_ ___   ___ ___ | |___
# |  ___/ '__/ _ \| __/ _ \ / __/ _ \| / __|
# | |   | | | (_) | || (_) | (_| (_) | \__ \
# |_|   |_|  \___/ \__\___/ \___\___/|_|___/


class TemplateApplicator(Protocol):
    def apply(self, logger) -> None: ...


class SupportsApplyTemplates(Protocol):
    def applicators(
        self,
        template_dir: str,
        src_dir: str,
        app_dir: str,
        program_name,
        app_name: str,
        logger,
    ) -> Sequence[TemplateApplicator]:
        """Support moving template files to their respective targets."""
        ...


class SupportsElmDependencies(Protocol):
    def install_elm_deps(
        self, working_dir: str, logger
    ) -> ExitSuccess[None] | ExitFailure[None, NPMError]:
        """Support installing an Elm package"""
        ...


class SupportsFlagLoader(Protocol):
    def load_flags(
        self,
        app_path: str,
        program_name: str,
        from_source: bool,
        watch_mode: bool,
        logger,
    ) -> ExitSuccess[Flags] | ExitFailure[None, Exception]:
        """Support loading a python flag module"""
        ...


class SupportsProgramHandlersCookieCutter(Protocol):
    def cookie_cutters(self, parent_dir: str, program_name: str) -> list[CookieCutter]:
        """Support adding a handlers JS module to a program"""
        ...


class SupportsModelCookieCutter(Protocol):
    def cookie_cutter(
        self,
        flags: Flags,
        app_name: str,
        program_name: str,
        src_path: str,
    ) -> CookieCutter:
        """Support generating a cookie cutter config for an Elm program Model"""
        ...


class SupportsProgramCookieCutter(Protocol):
    def cookie_cutters(
        self, app_name: str, program_name: str, src_path: str, version: str
    ) -> list[CookieCutter]:
        """Support generating a cookie cutter config for an Elm program"""
        ...


class ModelBuilder(
    SupportsModelCookieCutter, SupportsFlagLoader, SupportsApplyTemplates
):
    pass


class ProgramBuilder(
    SupportsProgramCookieCutter,
    SupportsApplyTemplates,
    SupportsElmDependencies,
):
    pass


class ProgramHandlersBuilder(
    SupportsProgramHandlersCookieCutter,
):
    pass


Lists = TypedDict("Lists", {"imports": list[str], "extras": list[str]})

EntrypointCookieExtra = TypedDict(
    "EntrypointCookieExtra",
    {
        "base_name": str,
        "base_path": str,
        "dir": str,
        "program_name": str,
        "scope": str,
        "view_name": str,
        "lists": Lists,
    },
)

HandlersCookieExtra = TypedDict(
    "HandlersCookieExtra",
    {
        "dir": str,
        "program_name": str,
    },
)


WidgetProgramCookieExtra = TypedDict(
    "WidgetProgramCookieExtra",
    {
        "tmp_dir": str,
        "program_name": str,
        "tag_file": str,
        "app_name": str,
        "scope": str,
        "view_name": str,
        "version": str,
    },
)
ModelCookieExtra = TypedDict(
    "ModelCookieExtra",
    {
        "program_name": str,
        "alias_type": str,
        "decoder_body": str,
        "tmp_dir": str,
        "module_path": str,
    },
)

ProgramCookieExtra = TypedDict(
    "ProgramCookieExtra",
    {
        "program_name": str,
        "view_name": str,
        "tmp_dir": str,
        "tag_file": str,
        "scope": str,
        "app_name": str,
        "version": str,
    },
)


def widget_cookie_cutter(
    app_name: str, src_path: str, program_name: WIDGET_NAMES_T, version: str
) -> CookieCutter:
    return CookieCutter[WidgetProgramCookieExtra](
        file_dir=os.path.dirname(__file__),
        output_dir=os.path.join(src_path, *STUFF_NAMESPACE),
        cookie_template_name="widget_templates",
        extra={
            "tmp_dir": "widgets",
            "program_name": program_name,
            "tag_file": tag_file_name(program_name),
            "app_name": app_name,
            "scope": widget_scope_name(app_name, program_name),
            "view_name": view_name(program_name),
            "version": version,
        },
        overwrite=True,
    )


def entrypoint_cookie_cutter(
    base_name: str,
    base_path: str,
    src_path: str,
    program_name: str,
    scope: str,
    view_prefix: str,
    imports: list[str] = [],
    extras: list[str] = [],
) -> CookieCutter:
    default_imports = ['import defo from "@icelab/defo";']
    return CookieCutter[EntrypointCookieExtra](
        file_dir=os.path.join(os.path.dirname(__file__), "cookiecutters"),
        output_dir=os.path.join(src_path, *STUFF_NAMESPACE),
        cookie_template_name="entrypoint_template",
        extra={
            "base_name": base_name,
            "base_path": base_path,
            "dir": "entrypoints",
            "program_name": program_name,
            "scope": scope,
            "view_name": view_prefix + view_name(program_name),
            "lists": {"imports": default_imports + imports, "extras": extras},
        },
        overwrite=True,
    )


def model_cookie_cutter(
    flags: Flags, program_name: str, tmp_dir: str, output_dir: str, module_path: str
):
    return CookieCutter[ModelCookieExtra](
        file_dir=os.path.dirname(__file__),
        output_dir=(output_dir),
        cookie_template_name="model_template",
        extra={
            "program_name": module_name(program_name),
            "tmp_dir": tmp_dir,
            "alias_type": flags.to_elm_parser_data()["alias_type"],
            "decoder_body": flags.to_elm_parser_data()["decoder_body"],
            "module_path": module_path,
        },
        overwrite=True,
    )


class ModelChoiceFieldWidgetGenerator(ProgramBuilder):
    """Generate all files for the ModelChoiceField widget"""

    def install_elm_deps(self, working_dir: str, logger):
        deps = [
            "elm-community/list-extra",
            "rtfeldman/elm-css",
            "Confidenceman02/elm-select",
        ]
        dep_string = ""
        for dep in deps:
            dep_string += f"\033[93m\t{dep}\033[0m\n"
        logger.write(
            f"""\n-- INSTALLING WIDGET DEPENDENCIES -------------------------------------- widget/ModelChoiceField
{dep_string}"""
        )
        args = ["elm-json", "install", "--yes"]
        args.extend(deps)
        effect = NPM(raise_err=True).command(working_dir, args)

        if effect.tag != "Success":
            return ExitFailure(None, effect.err)
        return ExitSuccess(None)

    def cookie_cutters(
        self, app_name: str, program_name: str, src_path: str, version: str
    ) -> list[CookieCutter]:
        return [
            widget_cookie_cutter(
                app_name, src_path, cast(WIDGET_NAMES_T, program_name), version
            ),
            entrypoint_cookie_cutter(
                base_name="Widgets.",
                base_path=f"Widgets{os.path.sep}",
                src_path=src_path,
                program_name=module_name(program_name),
                scope=widget_scope_name(app_name, program_name),
                view_prefix="widget",
            ),
        ]

    def applicators(
        self,
        template_dir: str,
        src_dir: str,
        app_dir: str,
        program_name: str,
        app_name: str,
        logger,
    ) -> Sequence[TemplateApplicator]:
        """Apply all generated templates to their respective targets"""

        logger.write(
            """\n-- GENERATING TEMPLATES ------------------------------------------------ widget/ModelChoiceField"""
        )
        return [
            # Program file
            TemplateCopyer(
                os.path.join(template_dir, module_name(program_name) + ".elmw"),
                os.path.join(
                    src_dir, "src", "Widgets", module_name(program_name) + ".elm"
                ),
            ),
            # Flags
            TemplateCopyer(
                os.path.join(template_dir, tag_file_name(program_name) + ".pyf"),
                os.path.join(
                    app_dir,
                    "flags",
                    "widgets",
                    tag_file_name(program_name) + ".py",
                ),
            ),
            # Tags
            TemplateCopyer(
                os.path.join(template_dir, f"{tag_file_name(program_name)}_tags.py"),
                os.path.join(
                    app_dir,
                    "templatetags",
                    f"{tag_file_name(program_name)}_widget_tags.py",
                ),
            ),
        ]


class ModelGenerator(ModelBuilder):
    def cookie_cutter(
        self, flags: Flags, app_name: str, program_name: str, src_path: str
    ) -> CookieCutter:
        return model_cookie_cutter(
            flags,
            program_name,
            STUFF_NAMESPACE[1],
            os.path.join(src_path, STUFF_NAMESPACE[0]),
            "Models",
        )

    def load_flags(
        self,
        app_path: str,
        program_name: str,
        from_source: bool,
        watch_mode: bool,
        logger,
    ) -> ExitSuccess[Flags] | ExitFailure[None, Exception]:
        if from_source:
            module = tag_file_name(program_name)
            module_path = os.path.join(
                app_path, "flags", tag_file_name(program_name) + ".py"
            )

            try:
                mod = FlagLoader.loader(module, module_path)
            except Exception as err:
                error_hint = ""
                if watch_mode:
                    error_hint = "\033[4m\033[1mNote\033[0m: I'll keep watching this module and will generate the model once the errors are fixed."
                logger.write(
                    f"""
    \033[91m-- FLAG MODULE ERROR -------------------------------------------------- command/generatemodel\033[0m

    I was trying to generate a model from this flags module:

        \033[93m{module_path}\033[0m

    But got the following error type:

        \033[93m{err.args[0]}\033[0m

    Make sure you fix the errors in \033[1m{module_path}\033[0m.

    {error_hint}

    """
                )
                return ExitFailure(None, err=err)

            return ExitSuccess(getattr(mod, program_name + "Flags"))
        else:
            # Return default flags for a new program
            return ExitSuccess(Flags(IntFlag()))  # type:ignore

    def applicators(
        self,
        template_dir: str,
        src_dir: str,
        app_dir: str,
        program_name: str,
        app_name: str,
        logger,
    ) -> Sequence[TemplateApplicator]:
        logger.write(
            f"""\n-- GENERATING MODEL ----------------------------------------------------- model/{module_name(program_name)}"""
        )
        return [
            TemplateCopyer(
                os.path.join(template_dir, module_name(program_name) + ".elmf"),
                os.path.join(
                    src_dir, "src", "Models", module_name(program_name) + ".elm"
                ),
            )
        ]


class ProgramGenerator(ProgramBuilder):
    """Generate an Elm program"""

    def install_elm_deps(
        self, working_dir: str, logger
    ) -> ExitSuccess[None] | ExitFailure[None, NPMError]:
        return ExitSuccess(None)

    def cookie_cutters(
        self, app_name: str, program_name: str, src_path: str, version: str
    ) -> list[CookieCutter]:
        cutters = []
        cutters.extend(
            [
                CookieCutter[ProgramCookieExtra](
                    file_dir=os.path.dirname(__file__),
                    output_dir=os.path.join(src_path, "elm-stuff"),
                    cookie_template_name="program_template",
                    extra={
                        "program_name": module_name(program_name),
                        "view_name": view_name(program_name),
                        "tmp_dir": STUFF_NAMESPACE[1],
                        "tag_file": tag_file_name(program_name),
                        "scope": scope_name(app_name, program_name),
                        "app_name": app_name,
                        "version": version,
                    },
                    overwrite=True,
                ),
                CookieCutter[EntrypointCookieExtra](
                    file_dir=os.path.join(os.path.dirname(__file__), "cookiecutters"),
                    output_dir=os.path.join(src_path, *STUFF_NAMESPACE),
                    cookie_template_name="entrypoint_template",
                    extra={
                        "base_name": "",
                        "base_path": "",
                        "dir": "entrypoints",
                        "program_name": program_name,
                        "scope": scope_name(app_name, program_name),
                        "view_name": "",
                        "lists": {"imports": [], "extras": []},
                    },
                    overwrite=True,
                ),
            ]
        )
        return cutters

    def applicators(
        self,
        template_dir: str,
        src_dir: str,
        app_dir: str,
        program_name,
        app_name: str,
        logger,
    ) -> Sequence[TemplateApplicator]:
        logger.write(
            f"""\n-- GENERATING TEMPLATES -------------------------------------------------------- {module_name(program_name)}"""
        )
        return [
            # Program file
            TemplateCopyer(
                os.path.join(template_dir, module_name(program_name) + ".elm"),
                os.path.join(src_dir, "src"),
            ),
            # Template tags
            TemplateCopyer(
                os.path.join(template_dir, tag_file_name(program_name) + "_tags.py"),
                os.path.join(app_dir, "templatetags"),
            ),
            # Flags
            TemplateCopyer(
                os.path.join(template_dir, tag_file_name(program_name) + ".pyf"),
                os.path.join(app_dir, "flags", tag_file_name(program_name) + ".py"),
            ),
        ]


class ProgramHandlersGenerator(ProgramHandlersBuilder):
    """
    Generates JS handlers for programs

    base_path: The base path from the 'static_src' directory not including the directory
    the handler file will be added to.

            If The handler is being created for a program in "src" then
            the base_path is [""]

            If the handler is being created for a program in src/Widgets then
            the base_path is ["src"]

    target_dir: The directory the handler will be added to

            If the handler is being added to the 'src' directory then
            the target_dir is 'src'

            If the handler is being added to the 'src/Widgets' directory then
            the target_dir is 'Widgets'
    """

    base_path: list[str]
    target_dir: str

    def __init__(self, base_path: list[str], target_dir: str) -> None:
        self.base_path = base_path
        self.target_dir = target_dir

    def cookie_cutters(self, parent_dir: str, program_name: str) -> list[CookieCutter]:
        """
        parent_dir: parent directory of the generated cookiecutters

            If the cookiecutter is creating a 'bar' directory inside a 'foo' directory (foo/bar), then
            the parent_dir is 'foo'.

        program_name: name of the program i.e. Main
        """
        return [
            CookieCutter[HandlersCookieExtra](
                file_dir=os.path.join(os.path.dirname(__file__), "cookiecutters"),
                output_dir=os.path.join(parent_dir, *self.base_path),
                cookie_template_name="handlers_template",
                extra=HandlersCookieExtra(
                    {"dir": self.target_dir, "program_name": program_name}
                ),
                overwrite=True,
            ),
        ]


class WidgetModelGenerator(ModelBuilder):
    """Generates models for widget programs"""

    def cookie_cutter(
        self, flags: Flags, app_name: str, program_name: str, src_path: str
    ) -> CookieCutter:
        return model_cookie_cutter(
            flags,
            program_name,
            "Models",
            os.path.join(src_path, *STUFF_NAMESPACE, "Widgets"),
            "Widgets.Models",
        )

    def load_flags(
        self,
        app_path: str,
        program_name: str,
        from_source: bool,
        watch_mode: bool,
        logger,
    ) -> ExitSuccess[Flags] | ExitFailure[None, Exception]:
        """
        Load a flag class for the given program

        app_path: The path of the djelm app
        program_name: The Elm program name
        from_source:
            True =  Load the flag class from a python module in the 'flags' directory
            False = Load a default flag class
        watch_mode:
            True = In a watch mode context
            False = Not in a watch mode context
        """
        if from_source:
            module = tag_file_name(program_name)
            module_path = os.path.join(
                app_path, "flags", "widgets", tag_file_name(program_name) + ".py"
            )

            try:
                mod = FlagLoader.loader(module, module_path)
            except Exception as err:
                error_hint = ""
                if watch_mode:
                    error_hint = "\033[4m\033[1mNote\033[0m: I'll keep watching this module and will generate the model once the errors are fixed."
                logger.write(
                    f"""
    \033[91m-- FLAG MODULE ERROR -------------------------------------------------- command/generatemodel\033[0m

    I was trying to generate a model from this flags module:

        \033[93m{module_path}\033[0m

    But got the following error type:

        \033[93m{err.args[0]}\033[0m

    Make sure you fix the errors in \033[1m{module_path}\033[0m.

    {error_hint}

    """
                )
                return ExitFailure(None, err=err)

            return ExitSuccess(getattr(mod, program_name + "Flags"))
        else:
            return ExitSuccess(Flags(WIDGET_NAME_TO_DEFAULT_FLAG[program_name]()))  # type:ignore

    def applicators(
        self,
        template_dir: str,
        src_dir: str,
        app_dir: str,
        program_name,
        app_name: str,
        logger,
    ) -> Sequence[TemplateApplicator]:
        logger.write(
            f"""\n-- GENERATING WIDGET MODEL --------------------------------------------- widget/{module_name(program_name)}"""
        )
        """Move the generated Elm model to their target location"""
        return [
            TemplateCopyer(
                os.path.join(template_dir, module_name(program_name) + ".elmf"),
                os.path.join(
                    src_dir,
                    "src",
                    "Widgets",
                    "Models",
                    module_name(program_name) + ".elm",
                ),
            )
        ]


class TemplateCopyer(TemplateApplicator):
    """Copy a template file to a given destination"""

    def __init__(self, src: str, destination: str, log: bool = True) -> None:
        self.src = src
        self.destination = destination
        self.log = log

    def apply(self, logger) -> None:
        shutil.copy(self.src, self.destination)
        if self.log:
            logger.write(f"\t\033[93m{self.destination}\033[0m")



import types


def loader(module: str, module_path: str) -> types.ModuleType:
    import importlib.machinery

    loader = importlib.machinery.SourceFileLoader(
        module,
        module_path,
    )

    mod = types.ModuleType(loader.name)

    loader.exec_module(mod)

    return mod



from dataclasses import dataclass

from djelm import get_config
from djelm.effect import ExitFailure, ExitSuccess
from djelm.subprocess import SubProcess


class NPMError(Exception):
    pass


@dataclass(slots=True)
class NPM:
    raise_err: bool = True

    def command(
        self, cwd: str, args: list[str]
    ) -> ExitSuccess[None] | ExitFailure[None, NPMError]:
        npm_bin_path = get_config("NODE_PACKAGE_MANAGER")
        try:
            process = SubProcess([npm_bin_path, *args], cwd, self.raise_err)
            process.open()
            return ExitSuccess(None)
        except OSError as err:
            return ExitFailure(
                None,
                NPMError(_to_npm_error(err, " ".join([npm_bin_path, *args]))),
            )


def _to_npm_error(err: Exception, command: str):
    return f"""
\033[91m-- PACKAGE MANAGER ERROR ---------------------------------------------------------------------------------------------------------- command/npm\033[0m

Im trying to run this command:

    \033[93m{command}\033[0m

But it raised this error:

    \033[93m{err}\033[0m

\033[4m\033[1mHint\033[0m: If I don't see a \033[1mNODE_PACKAGE_MANAGER\033[0m variable in \033[1msettings.py\033[0m I will try to use \033[1mpnpm\033[0m by default.

Check out <https://github.com/Confidenceman02/django-elm?tab=readme-ov-file#npm-command> for more information."""



from dataclasses import dataclass
import typing

from pydantic import BaseModel


class Flag:
    pass


@dataclass(slots=True)
class StringFlag(Flag):
    """Flag for the Elm String primitive

    literal :
        Will match against the passed string literal.

        Generated decoders will also express the string literal
    """

    literal: str | None = None


@dataclass(slots=True)
class IntFlag(Flag):
    """Flag for the Elm Int primitive"""


@dataclass(slots=True)
class FloatFlag(Flag):
    """Flag for the Elm Float primitive"""


@dataclass(slots=True)
class BoolFlag(Flag):
    """Flag for the Elm Bool primitive"""


@dataclass(slots=True)
class NullableFlag(Flag):
    """Flag for the Elm Maybe monad"""

    obj: Flag


@dataclass(slots=True)
class ListFlag(Flag):
    """Flag for the Elm List primitive"""

    obj: Flag


@dataclass(slots=True)
class ObjectFlag(Flag):
    """Flag for the Elm {} primitive"""

    obj: typing.Dict[str, Flag]


@dataclass(slots=True)
class CustomTypeFlag(Flag):
    """
    A Flag for an Elm custom type.

    <https://guide.elm-lang.org/types/custom_types>

    variants = A list of tuples that specify the a discriminator and flag.

    i.e.
            FlagModel = Flags(CustomTypeFlag(variants=[("Custom1",StringFLag()), ("Custom2",IntFlag())]))

            Turns into the the Elm type:

            type SomeCustomType
                = Custom1 String
                | Custom2 Int

            Then we parse a valid data structure:

            FlagModel.parse("Hi there")
    """

    variants: list[tuple[str, Flag]]


FlagsObject = dict[str, "PrimitiveFlag"]
FlagsList = list["PrimitiveFlag"]
FlagsNullable = typing.Union[type[str], type[int], type[float], type[bool], type[None]]
PrimitiveObjectFlagType = (
    type[str]
    | type[int]
    | type[float]
    | type[bool]
    | type[BaseModel]
    | type[list]
    | FlagsNullable
)
PrimitiveFlag = (
    str | int | float | bool | FlagsObject | FlagsList | FlagsNullable | None
)



import argparse

from django.core.management.base import LabelCommand

from ...strategy import (
    AddProgramHandlersStrategy,
    AddProgramStrategy,
    AddWidgetStrategy,
    CompileStrategy,
    CreateStrategy,
    ElmStrategy,
    FindProgramsStrategy,
    GenerateModelStrategy,
    GenerateModelsStrategy,
    ListStrategy,
    ListWidgetsStrategy,
    NpmStrategy,
    Strategy,
    WatchStrategy,
)
from ...validate import Validations


class Command(LabelCommand):
    help = "Runs elm commands"
    missing_args_message = """
Command arguments are missing, please add one of the following:
  create <app-name> - to create a djelm app
  addprogram <app-name> <program-name> - create an Elm program called <program-name> in the <app-name> app
  addwidget <app-name> <widget-name> - create a widget program in the <app-name> app
  watch <app-name> - will watch the app's src file for Elm code changes and compile
  npm <app-name> [args].. - call your designated NODE_PACKAGE_MANAGER with [args]
  elm <app-name> [args].. - call your designated ELM_BIN_PATH with [args]
  generatemodel <app-name> [args].. - generate a model for the existing program <program-name> in the <app-name> app
  generatemodel <app-name> - generate a model for all programs in the <app-name> app
  list - to list all your djelm apps
  listwidgets - to list all supported widget programs
  findprograms - to list all Elm programs in src/
  compile <app-name> - to compile all your elm programs in the given <app-name> app
  compilebuild <app-name> - to compile all your elm programs with a production level build in the given <app-name> app
Usage example:
  python manage.py djelm create djelm_app
  python manage.py djelm addprogram djelm_app MyElmProgram
  python manage.py djelm addwidget djelm_app ModelChoiceField
  python manage.py djelm watch djelm_app
  python manage.py djelm npm djelm_app install
  python manage.py djelm elm djelm_app install <elm-package>
  python manage.py djelm generatemodel djelm_app MyElmProgram
  python manage.py djelm generatemodels djelm_app
  python manage.py djelm list
  python manage.py djelm listwidgets
  python manage.py djelm findprograms djelm_app
  python manage.py djelm compile djelm_app
  python manage.py djelm compilebuild djelm_app
"""
    validate = None
    strategy: (
        CreateStrategy
        | AddProgramHandlersStrategy
        | AddProgramStrategy
        | AddWidgetStrategy
        | CompileStrategy
        | ElmStrategy
        | FindProgramsStrategy
        | GenerateModelStrategy
        | GenerateModelsStrategy
        | ListStrategy
        | ListWidgetsStrategy
        | NpmStrategy
        | WatchStrategy
    )

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.validate = Validations()

    def add_arguments(self, parser):
        _, unknown = parser.parse_known_args()

        if 1 < len(unknown) and unknown[1] in ["npm", "elm"]:
            parser.add_argument("strategy")
            parser.add_argument("app")
            parser.add_argument("rest", nargs=argparse.REMAINDER)

        else:
            parser.add_argument("--no-deps", action="store_true")
            super(Command, self).add_arguments(parser)

    def handle(self, *labels, **options):  # type:ignore
        if len(labels) == 0:
            # We are dealing with a npm strategy when there are no labels
            return self.handle_labels(
                *(options["strategy"], options["app"], *options["rest"])
            )
        return self.handle_labels(*labels, **options)

    def handle_labels(self, *labels, **options):
        self.strategy = Strategy().create(list(labels), options)
        self.run_strategy()

    def run_strategy(self):
        self.strategy.run(self.stdout)
